//////////////////////////// RCS COMMENTS: /////////////////////////////
//  HISTORY LOG:
//  ------------
//  $Log: ast.cc,v $
// Revision 1.15  1998/12/07  22:53:46  madings
// Checking in multiple error finder for the time being.
// Not done yet but someone needs to check out what I have
// so far...
//
// Revision 1.14  1998/11/20  06:46:42  madings
// Bugfixing.
//
// Revision 1.13  1998/10/08  03:45:29  madings
// Two changes:
//    1 - Most signifigant change is that the library now has
//    doc++ comments in it.
//    2 - Also, changed so that when using the copy constructor,
//    it remembers line numbers and column numbers in the newly
//    copied tree (I hope, this is untested.)
//
// Revision 1.12  1998/09/04  07:02:00  madings
// There were several bug fixes with this version.
//
// However, the erasing probably still needs some work.
//
// Revision 1.11  1998/08/29  08:45:55  madings
// This one was a rather large change.  I haven't finished
// documenting and testing all of it, so expect another
// revision or two soon.
//
// This revision creates a new way of dealing with loops.  Now
// the caller can use the loops like 2-D vectors, but with some
// extra magic behind the scenes to ensure that no insertions or
// deletions that can leave the tree in an invalid state will
// be allowed.  (For example, you cannot remove a value from the
// loop without also removing the tagname for it and all the other
// values in that column.)
//
// Also in this revision is a fix to the previous revision's method
// of making List and ASTlist be wrappers around STL vector.  The
// previous version of the library broke when the STL library decided
// to move the vector to a new location in memory (realloc), and the
// myCurrent interator became invalid.  Now myCurrent is an integer
// index offset instead, so it moves with the vector.
//
// Also in this revision are some fixes to a longstanding problem that
// made the library incapable of parsing nested loops unless the nested
// loops were fully populated.  (In other words if an inner loop for any
// row was null, because the input file simply has a "stop_" and no
// data for the inner loop, it used to crash on that.)  Now it works and
// empty inner loops are represented to the caller as null pointers (so
// don't blindly follow inner loop pointers that are handed to you - check
// them first.)
//
// Note that this marks my first attempt to make a class that implements
// the STL vector interface, so I expect to find some bugs to fix in the
// next revision or two.  There's some strange C++ "magic" going on here
// to get this to work.
//
// Revision 1.10  1998/08/18  01:00:02  madings
// Got rid of the "astoper.cc" and "interface.cc" files and appended
// their contents to "ast.cc".  There was no modular reason for them
// to be seperate.
//
// Revision 1.9  1998/08/14  00:01:13  madings
// Added ability to define which loops are to be output many-rows to the
// line, and the ability to specifcy that a loop's values should be printed
// up against the margin and not indented.
//
// Revision 1.8  1998/07/17  01:41:25  madings
// Two changes related to Unparse() this revision:
//   1 - Added the ability to programtically insert comments with
//           the get/setPreComment functions.
//   2 - Make Unparse() somewhat more sophisticated to handle formatting
// 	  values so they line up with each other.
//
// Revision 1.7  1998/06/02  06:21:58  madings
// Made the yacc code that keeps track of lineNum/colNum a bit more
// sophisticated so that it shows the locations of the beginnings
// of the objects, not the ends of them.  Also gave the base class
// 'ASTnode' a copy constructor that will copy the increasing number
// of variables that are being stored in astnode, like the lineNum/colNum.
//
// Revision 1.6  1998/05/29  06:51:12  madings
// Added the ability to search for all objects of a particular AST type.
// (i.e. all SaveFrameNodes, or all DataItemNodes, etc...)
// See the method called "searchForType()", which was added to do this.
//
// Revision 1.5  1998/05/25  07:15:14  madings
// Added preliminary support for keeping track of line numbers and column
// numbers from the parse.  This way, programs reading the data can give
// the user actual line/column number references to things found in the
// STAR file.
//
// Revision 1.4  1998/05/21  22:20:44  madings
// Made some minor changes from string references to string-by-value
// in a few function templates.
//
// Revision 1.3  1998/05/08  20:27:23  informix
// Made a change to handle the C++ STL instead of gcc's libg++.
// Now all strings are of type 'string' instead of 'String' (notice
// the uppercase/lowercase change.)  Applications using this library
// will have to change also, unfortunately, to use the new string style.
//
// Revision 1.2  1998/02/21  01:12:41  madings
// changed the text of the "Error transforming loop" message to
// make more sense to a human.
//
// Revision 1.1.1.1  1998/02/12  20:59:00  informix
// re-creation of starlib project, without sb_lib stuff.  sb_lib has
// been moved to it's own project.
//
// Revision 1.1.1.1  1997/11/07  01:49:05  madings
// First check-in of starlib and sb_lib in library form.
//
// Revision 1.7  1997/10/25  02:59:24  madings
//  2 Things in this check-in:
//    -1- "Parallel copy".  Addded a new way to do a copy constructor
//        such that it will copy with a parallel link back to the
//        original tree it was copied from.  Any node in either tree
//        can refer to the corresponding node in the new tree.
//        See "parallel.html" for more detail.
//
//    -2- Bug fixes to copy constructors that did not properly deep-copy
//        all the way down into the LoopIter values.
//
// Revision 1.6  1997/10/10  18:34:19  madings
// Changed DataValueNode so that it now no longer has subclasses for
// single-quote-delimited strings, double-quote delimited strings,
// semicolon-delimited strings, and so on.  Now there is only one kind
// of DataValueNode, and it uses a flag to decide on the delimiter type.
// This also allows for the creation of methods to change a DataValueNode
// in place, which was not previously possible when different kinds of
// values had to be stored in different kinds of classes.
//
// Revision 1.5  1997/10/08  20:23:17  madings
// Added "searchForTypeByTag()" and "searchForTypeByTagValue()".
//
// (Also fixed bug where it would not find saveframes.)
//
// Revision 1.4  1997/10/01  22:35:02  bsri
// This version of starlib split main.cc into separate files (which contain
// the transformation functions - one in each file).
//
// Revision 1.2  1997/09/16  22:03:17  madings
// Changed searchByTag() and searchByTagValue() so that they will
// return lists instead of returning single hits, and extended
// them to levels of the star tree above "saveframe" so it is now
// possible to use them on the whole star file.
// (Also changed the history log comments of several header files
// to use slash-star comment style instead of '//' style.)
//
// Revision 1.1.1.1  1997/09/12  22:54:39  madings
// Restarting CVS tree because I lost the CVS dir.
//
//Revision 1.6  1997/07/30  21:34:26  madings
//check-in before switching to cvs
//
//Revision 1.5  1997/07/11  22:40:03  madings
//Finally seems to work - just waiting for
//new dictionaries.
//
//Revision 1.4  1997/06/13  22:52:44  madings
//Checking in before bringing in sridevi code.
//
//Revision 1.3  1997/06/04  18:57:57  madings
//The RCS ID variables were missing the trailing $, so I
//changed them across all these files.
//
//Revision 1.2  1997/06/04  18:52:43  madings
//I Added the RCS log and ID variables.
//
//
////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <assert.h>
#include "ast.h"

#ifdef IS_GCC
#   pragma implementation "ast.h"
#endif

ASTnode::ASTnode( void )
{
    parent     = (ASTnode*)NULL;
    peer       = (ASTnode*)NULL;
    amDeleting = false;
    lineNum    = -1;
    colNum     = -1;
    preComment = string("");

#ifdef DEBUG_AST_CONST
    cerr << "DEBUG_AST_CONST: new object's ptr = " << (void*)this <<
	", type = " << myType() << endl;
#endif
}

ASTnode::ASTnode( const ASTnode &copyFromMe )
{
    copyFrom( copyFromMe );
}

void ASTnode::copyFrom( const ASTnode &copyFromMe )
{
    parent     = copyFromMe.parent;
    peer       = copyFromMe.peer;
    amDeleting = copyFromMe.amDeleting;
    lineNum    = copyFromMe.lineNum;
    colNum     = copyFromMe.colNum;
}

void ASTnode::setPeer( ASTnode *to_this )
{
    peer = to_this;
}

void ASTnode::setParent( ASTnode *p )
{
    parent = p;
    // If line number and column number are not set yet, then
    // set them to be the same as the parent's.
    if( lineNum < 0  &&  colNum < 0  && parent != NULL )
    {
	lineNum = parent->getLineNum();
	colNum = parent->getColNum();
    }
}

ASTnode * ASTnode::myParallelCopy( void )
{
    return peer;
}

List<ASTnode*> * ASTnode::searchByTag( string & )
{
    // The default behavior if the coder forgets to override
    // this function in a derivative of ASTnode.  Hopefully
    // this never gets executed.  It is just here to get
    // it to compile without complaint for those derivatives
    // of ASTnode for which searchByTag() makes no sense.
    return new List<ASTnode*>;  // return an empty list.
}

List<ASTnode*> * ASTnode::searchByTag( char *searchFor )
{
    // Just a wrapper around the 'real' searchByTag call,
    // which expects a string object.  Note that since
    // these are virtual functions, you never need to
    // redefine this function for any child of ASTnode.
    // this definition here should handle all uses of
    // the (char*) version of this function, for any
    // subclass of ASTnode.
    List<ASTnode*> *retVal;
    string  *searchString = new string(searchFor);

    retVal = searchByTag( *searchString );

    delete searchString;
    return retVal;
}

List<ASTnode*> * ASTnode::searchByTagValue( string &, string & )
{
    // The default behavior if the coder forgets to override
    // this function in a derivative of ASTnode.  Hopefully
    // this never gets executed.  It is just here to get
    // it to compile without complaint for those derivatives
    // of ASTnode for which searchByTagValue() makes no sense.
    return new List<ASTnode*> ;  // return an empty list
}

List<ASTnode*> * ASTnode::searchByTagValue( string &tag,
				     char   *value )
{
    // Just a wrapper around the first version of this function.
    // ---------------------------------------------------------
    List<ASTnode*> *retVal;
    string  *valueStr = new string(value);

    retVal = searchByTagValue( tag, *valueStr );

    delete valueStr;
    return retVal;
}

List<ASTnode*> * ASTnode::searchByTagValue( char   *tag,
				     string &value )
{
    // Just a wrapper around the first version of this function.
    // ---------------------------------------------------------
    List<ASTnode*> *retVal;
    string  *tagStr = new string(tag);

    retVal = searchByTagValue( *tagStr, value );

    delete tagStr;
    return retVal;
}

List<ASTnode*> * ASTnode::searchByTagValue( char   *tag,
				     char   *value )
{
    // Just a wrapper around the first version of this function.
    // ---------------------------------------------------------
    List<ASTnode*> *retVal;
    string  *tagStr = new string(tag);
    string  *valueStr = new string(value);

    retVal = searchByTagValue( *tagStr, *valueStr );

    delete valueStr;
    delete tagStr;
    return retVal;
}

List<ASTnode*> *ASTnode::searchForTypeByTag(
	  ASTtype  type,      // type to search for.
	  string &tag         // tag to search for.
	  )
{
    ASTnode        *remember;
    List<ASTnode*> *retVal = searchByTag( tag );

    for( retVal->Reset() ; ! retVal->AtEnd() ; retVal->Next() )
    {
	ASTnode *thisNode = retVal->Current();
	while( 	thisNode != (ASTnode*)NULL  &&  ! thisNode->isOfType( type ) )
	    thisNode = thisNode->myParent();
	remember = retVal->Current();
	if( thisNode == (ASTnode*)NULL )
	    retVal->RemoveCurrent();
	else
	    retVal->AlterCurrent( thisNode );
	retVal->SeekTo( remember );
    }

    return retVal;
}

List<ASTnode*> *ASTnode::searchForTypeByTag(
	  ASTtype  type,      // type to search for.
	  char *tag           // tag to search for.
	  )
{
    string tagStr( tag );
    return searchForTypeByTag( type, tagStr );
}

List<ASTnode*> *ASTnode::searchForTypeByTagValue(
	  ASTtype  type,   // type to search for.
	  string   &tag,   // Look for this tag...
	  string   &value  // Where it has this value.
	  )
{
    ASTnode        *remember;
    List<ASTnode*> *retVal = searchByTagValue( tag, value );

    for( retVal->Reset() ; ! retVal->AtEnd() ; retVal->Next() )
    {
	ASTnode *thisNode = retVal->Current();
	while( 	thisNode != (ASTnode*)NULL  &&  ! thisNode->isOfType( type ) )
	    thisNode = thisNode->myParent();
	remember = retVal->Current();
	if( thisNode == (ASTnode*)NULL )
	    retVal->RemoveCurrent();
	else
	    retVal->AlterCurrent( thisNode );
	retVal->SeekTo( remember );
    }

    return retVal;
}

List<ASTnode*> *ASTnode::searchForTypeByTagValue(
	  ASTtype  type,  // type to search for.
	  string    &tag, // Look for this tag...
	  char     *value // Where it has this value.
	  )
{
    string valStr( value );
    return searchForTypeByTagValue( type, tag, valStr );
}

List<ASTnode*> *ASTnode::searchForTypeByTagValue(
	  ASTtype  type,  // type to search for.
	  char     *tag,  // Look for this tag...
	  string   &value // Where it has this value.
	  )
{
    string tagStr( tag );
    return searchForTypeByTagValue( type, tagStr, value );
}

List<ASTnode*> *ASTnode::searchForTypeByTagValue(
	  ASTtype  type, // type to search for.
	  char    *tag,  // Look for this tag...
	  char    *value // Where it has this value.
	  )
{
    string tagStr( tag );
    string valStr( value );
    return searchForTypeByTagValue( type, tagStr, valStr );
}

List<ASTnode*> * ASTnode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*>  *retVal;

    delim;  // get the compiler to shut up about unused arguments.

    // The default behavior if the coder forgets to override
    // this function in a derivative of ASTnode.
    // 
    // The default behavior is to simply return the current
    // node if the type matches.   This is the default behavior
    // that should be appropriate for any leaf node in the tree.
    // To nodes that are not leaves, the algorithm for walking
    // through the children has to be customized for each type
    // of ASTnode.  Thus, only those ASTnode's that are capable
    // of being non-leaf nodes will need to override this function.
    // (All they have to do in that case is call the seachForType()
    // of each of their children.)
    //
    retVal = new List<ASTnode*>;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    return retVal ;
}

bool ASTnode::removeChild( ASTnode *child )
{
    if( unlinkChild( child ) )
    {   delete child;
	return true;
    }
    else
	return false;
}

bool ASTnode::unlinkChild( ASTnode * )
{
    return false;  // do nothing by default.  Needs to be overridden.
}

bool ASTnode::removeMe( void )
{
    if( myParent() != (ASTnode*)NULL )
	return myParent()->removeChild(this);
    else
	return false;
}

bool ASTnode::unlinkMe( void )
{
    if( myParent() != (ASTnode*)NULL )
    {
	// When the parent deliberately deletes a child, that child turns
	// around and tells the parent to unlink itself.  But if the
	// parent is already in its destructor, then this is a Bad Idea.
	if( myParent()->amDeleting )
	{   return true;
	}
        else
	{   return myParent()->unlinkChild(this);
	}
    }
    else
    {   return false;
    }
}

ASTnode::~ASTnode( void )
{
#ifdef DEBUG_AST_DEST
    cerr << "DEBUG_AST_DEST: Removing following object: " << (void*)this <<
	", type = " << myType() << endl;
#endif
    amDeleting = true;

    // Unlink me from my parallel peer if I am
    // about to die:
    ASTnode *otherOne = myParallelCopy();
    if( otherOne != NULL )
	otherOne->setPeer( NULL );
    amDeleting = false;
}

HeadingNode::HeadingNode(HeadingNode & h)
{
    (void)copyFrom( h, false );
}

HeadingNode::HeadingNode( bool link, HeadingNode & h)
{
    (void)copyFrom( h, link );
}

bool HeadingNode::copyFrom( HeadingNode &fromThis, bool linkWith )
{
    ASTnode::copyFrom( fromThis );
    myHeading = fromThis.myHeading;

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

HeadingNode::HeadingNode(const string & h) : myHeading(h) {}

void HeadingNode::changeName(string &name) 
{
    myHeading = name;
}

DataListNode::DataListNode(ASTlist<DataNode *> *L) : myKids(L)
{
    myKids->setParent( this );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	myKids->Current()->setParent( this );
}

bool DataListNode::copyFrom( DataListNode &fromThis, bool linkWith )
{
    DataListNode &L    = fromThis;
    DataNode *remember = (L.myKids->AtEnd()) ? 
			(DataNode*)NULL : L.myKids->Current();

    ASTnode::copyFrom( fromThis );
    myKids = new ASTlist<DataNode *>;
    myKids->setParent( this );
    for( L.myKids->Reset() ; ! L.myKids->AtEnd() ; L.myKids->Next() )
    {
	switch( L.myKids->Current()->myType() )
	{
	    case ASTnode::DATAITEMNODE:
		myKids->AddToEnd( new DataItemNode( linkWith, *( (DataItemNode*)L.myKids->Current() ) )  );
		break;
	    case ASTnode::DATALOOPNODE:
		myKids->AddToEnd( new DataLoopNode( linkWith, *( (DataLoopNode*)L.myKids->Current() ) )  );
		break;
	    case ASTnode::SAVEFRAMENODE: 
		myKids->AddToEnd( new SaveFrameNode( linkWith, *( (SaveFrameNode*)L.myKids->Current() ) )  );
		break;
	    default:   // This is here to get "gcc -Wall" to shut up.
		break;
	}
    }
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	myKids->Current()->setParent( this );
    (void)L.myKids->SeekTo( remember );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

DataListNode::DataListNode(DataListNode &L)
{
    (void)copyFrom( L, false );
}

DataListNode::DataListNode(bool link, DataListNode &L)
{
    (void)copyFrom( L, link );
}

DataListNode::~DataListNode()
{
    amDeleting = true;
    unlinkMe();
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        if( myKids->Current() != NULL )
 	    delete myKids->Current();

    if( myKids != NULL )
        delete myKids;
    amDeleting = false;
}

List<ASTnode*> * DataListNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *someHits;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	someHits = myKids->Current()->searchByTag( searchFor );
	if( someHits->Length() > 0 )
	    retVal->AddToEnd( *someHits );
	delete someHits;
    }

    return retVal;
}

List<ASTnode*> * DataListNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	matches = myKids->Current()->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    return retVal;
}

List<ASTnode*> * DataListNode::searchByTagValue( string &tag, string &value )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *someHits;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	someHits = myKids->Current()->searchByTagValue( tag, value );
	if( someHits->Length() > 0 )
	    retVal->AddToEnd( *someHits );
	delete someHits;
    }

    return retVal;
}

bool DataListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

int DataListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	// Loops are weird.  They don't count.
	if( ! myKids->Current()->isOfType( ASTnode::DATALOOPNODE )  )
	{
	    curLen = myKids->Current()->myLongestStr();
	    if( curLen > retVal )
		retVal = curLen;
	}
    }
    return retVal;
}

///////////////////////////////////////////////////////////////////////
//       myType()
// The compiler I was using has trouble with virtual inlined
// functions not behaving like a virtual function should, so I had to
// make this be not inlined for them to work, even though they are
// very simple:
//////////////////////////////////////////////////////////////////////

ASTnode::ASTtype ASTnode::myType( void )
{ return ASTNODE; }
ASTnode::ASTtype BlockNode::myType( void )
{ return BLOCKNODE; }
ASTnode::ASTtype DataBlockNode::myType( void )
{ return DATABLOCKNODE; }
ASTnode::ASTtype DataHeadingNode::myType( void )
{ return DATAHEADINGNODE; }
ASTnode::ASTtype DataItemNode::myType( void )
{ return DATAITEMNODE; }
ASTnode::ASTtype DataListNode::myType( void )
{ return DATALISTNODE; }
ASTnode::ASTtype DataLoopNameListNode::myType( void )
{ return DATALOOPNAMELISTNODE; }
ASTnode::ASTtype DataLoopValListNode::myType( void )
{ return DATALOOPVALLISTNODE; }
ASTnode::ASTtype LoopValListNode::myType( void )
{ return LOOPVALLISTNODE; }
ASTnode::ASTtype DataLoopNode::myType( void )
{ return DATALOOPNODE; }
ASTnode::ASTtype DataNameNode::myType( void )
{ return DATANAMENODE; }
ASTnode::ASTtype DataNode::myType( void )
{ return DATANODE; }
ASTnode::ASTtype DataValueNode::myType( void )
{ return DATAVALUENODE; }
ASTnode::ASTtype GlobalBlockNode::myType( void )
{ return GLOBALBLOCKNODE; }
ASTnode::ASTtype GlobalHeadingNode::myType( void )
{ return GLOBALHEADINGNODE; }
ASTnode::ASTtype HeadingNode::myType( void )
{ return HEADINGNODE; }
ASTnode::ASTtype LoopRowNode::myType( void )
{ return LOOPROWNODE; }
ASTnode::ASTtype LoopNameListNode::myType( void )
{ return LOOPNAMELISTNODE; }
ASTnode::ASTtype LoopTableNode::myType( void )
{ return LOOPTABLENODE; }
ASTnode::ASTtype SaveFrameListNode::myType( void )
{ return SAVEFRAMELISTNODE; }
ASTnode::ASTtype SaveFrameNode::myType( void )
{ return SAVEFRAMENODE; }
ASTnode::ASTtype SaveHeadingNode::myType( void )
{ return SAVEHEADINGNODE; }
ASTnode::ASTtype StarListNode::myType( void )
{ return STARLISTNODE; }
ASTnode::ASTtype StarFileNode::myType( void )
{ return STARFILENODE; }

/////////////////////////////////
//       isOfType()
/////////////////////////////////
// All of the various isOfType() functions have been grouped together here:

bool ASTnode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool HeadingNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool StarFileNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataListNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool BlockNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool StarListNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool GlobalBlockNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataBlockNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataNameNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataValueNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataItemNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool LoopNameListNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataLoopNameListNode ::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataLoopValListNode ::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool LoopValListNode ::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool LoopRowNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool LoopTableNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataLoopNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool GlobalHeadingNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool DataHeadingNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool SaveHeadingNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool SaveFrameListNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}
bool SaveFrameNode::isOfType( ASTtype T )
{
    return NotVirtualIsOfType(T);
}

  //////////////////////////////////////////////////////////////////////////
 /////////////////// NotVirtualIsOfType functions: ////////////////////////
//////////////////////////////////////////////////////////////////////////
bool ASTnode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == ASTNODE);
}
bool HeadingNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == HEADINGNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATANODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool StarFileNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == STARFILENODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATALISTNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool BlockNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == BLOCKNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool StarListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == STARLISTNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool GlobalBlockNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == GLOBALBLOCKNODE) ? true : ((BlockNode*)this)->NotVirtualIsOfType(T);
}
bool DataBlockNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATABLOCKNODE) ? true : ((BlockNode*)this)->NotVirtualIsOfType(T);
}
bool DataNameNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATANAMENODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataValueNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATAVALUENODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataItemNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATAITEMNODE) ? true : ((DataNode*)this)->NotVirtualIsOfType(T);
}
bool LoopNameListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == LOOPNAMELISTNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataLoopNameListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATALOOPNAMELISTNODE ) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataLoopValListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATALOOPVALLISTNODE ) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool LoopValListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == LOOPVALLISTNODE ) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool LoopRowNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == LOOPROWNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool LoopTableNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == LOOPTABLENODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool DataLoopNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATALOOPNODE) ? true : ((DataNode*)this)->NotVirtualIsOfType(T);
}
bool GlobalHeadingNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == GLOBALHEADINGNODE) ? true : ((HeadingNode*)this)->NotVirtualIsOfType(T);
}
bool DataHeadingNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == DATAHEADINGNODE) ? true : ((HeadingNode*)this)->NotVirtualIsOfType(T);
}
bool SaveHeadingNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == SAVEHEADINGNODE) ? true : ((HeadingNode*)this)->NotVirtualIsOfType(T);
}
bool SaveFrameListNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == SAVEFRAMELISTNODE) ? true : ((ASTnode*)this)->NotVirtualIsOfType(T);
}
bool SaveFrameNode::NotVirtualIsOfType( ASTtype T )
{
    return  (T == SAVEFRAMENODE) ? true : ((DataNode*)this)->NotVirtualIsOfType(T);
}
DataListNode::DataListNode()
{
    myKids = new ASTlist<DataNode *>;
    myKids->setParent( this );
}

BlockNode::BlockNode( BlockNode &n )
{
    (void)copyFrom( n, false );
}

BlockNode::BlockNode( bool link, BlockNode &n )
{
    (void)copyFrom( n, link );
}

bool BlockNode::copyFrom( BlockNode &fromThis, bool linkWith )
{
    ASTnode::copyFrom( fromThis );
    myDataList = new DataListNode( linkWith, *(fromThis.myDataList) );
    myHeading = new HeadingNode( linkWith, *(fromThis.myHeading) );
    myHeading->setParent(this);
    myDataList->setParent(this);

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

BlockNode::BlockNode(HeadingNode *h, DataListNode *n) 
:
 myDataList(n), myHeading(h)
{
    myHeading->setParent(this);
    myDataList->setParent(this);
}

BlockNode::BlockNode(const string & name)
{
  myHeading = new HeadingNode(name);
  myDataList = new DataListNode;
  myHeading->setParent(this);
  myDataList->setParent(this);
}

BlockNode::~BlockNode()
{
    amDeleting = true;
    unlinkMe();
    if( myDataList != NULL )
	delete myDataList;
    if( myHeading != NULL )
	delete myHeading;
    amDeleting = false;
}

List<ASTnode*> * BlockNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;
    
    if( myHeading != (ASTnode*)NULL )
    {
	if( myHeading->myName() == searchFor )
	    retVal->AddToEnd( this );
    }

    if( myDataList != (ASTnode*)NULL )
    {
	foundSome = myDataList->searchByTag( searchFor );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * BlockNode::searchForType( 
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;
    
    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( myHeading != (ASTnode*)NULL )
    {
	if( myHeading->isOfType( type ) )
	    retVal->AddToEnd( myHeading );
    }

    if( myDataList != (ASTnode*)NULL )
    {
	foundSome = myDataList->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * BlockNode::searchByTagValue( string &tag, string &value )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;
    
    if( myDataList != (ASTnode*)NULL )
    {
	foundSome = myDataList->searchByTagValue( tag, value );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

bool BlockNode::unlinkChild( ASTnode *child )
{
    if( myDataList == (DataListNode*)child )
    {	myDataList = (DataListNode*)NULL;
	return true;
    }
    else if( myHeading == (HeadingNode*)child )
    {   myHeading = (HeadingNode*)NULL;
	return true;
    }
    else
	return false;
}

int BlockNode::myLongestStr( void )
{
    int headLength = myHeading->myLongestStr();
    int listLength = myDataList->myLongestStr();

    // return whichever is greater:
    return ( headLength > listLength ) ? headLength : listLength ;
}

StarListNode::StarListNode( StarListNode &L)
{
    (void)copyFrom( L, false );
}

StarListNode::StarListNode( bool link, StarListNode &L)
{
    (void)copyFrom( L, link );
}

bool StarListNode::copyFrom( StarListNode &fromThis, bool linkWith )
{
    StarListNode &L = fromThis;

    BlockNode *remember = (L.myKids->AtEnd()) ? 
			    (BlockNode*)NULL : L.myKids->Current();
    ASTnode::copyFrom( fromThis );
    myKids = new ASTlist<BlockNode *>;
    myKids->setParent(this);
    for(   L.myKids->Reset() ; ! L.myKids->AtEnd() ; L.myKids->Next()  )
    {
	switch( L.myKids->Current()->myType() )
	{
	    case ASTnode::BLOCKNODE:
		myKids->AddToEnd( new BlockNode( linkWith, *(L.myKids->Current()) ) );
		break;
	    case ASTnode::DATABLOCKNODE:
		myKids->AddToEnd( new DataBlockNode( linkWith, *(DataBlockNode*)(L.myKids->Current()) ) );
		break;
	    case ASTnode::GLOBALBLOCKNODE:
		myKids->AddToEnd( new GlobalBlockNode( linkWith, *(GlobalBlockNode*)(L.myKids->Current()) ) );
		break;
	    default:   // Just here to keep gcc -Wall happy.
		break;
	}
    }
    for(   myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next()  )
    {   myKids->Current()->setParent(this);
    }
    (void)L.myKids->SeekTo( remember );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

StarListNode::StarListNode(ASTlist<BlockNode *> *L) : myKids(L)
{
    myKids->setParent(this);
    for(   myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next()  )
    {   myKids->Current()->setParent(this);
    }
}

StarListNode::StarListNode()
{
    myKids = new ASTlist<BlockNode *>;
    myKids->setParent(this);
    for(   myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next()  )
    {   myKids->Current()->setParent(this);
    }
}

StarListNode::~StarListNode()
{
    amDeleting = true;
    unlinkMe();
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	if( myKids->Current() != NULL )
	    delete myKids->Current();

    if( myKids != NULL )
	delete myKids;
    amDeleting = false;
}

List<ASTnode*> * StarListNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchByTag( searchFor );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * StarListNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * StarListNode::searchByTagValue( string &tag, string &value )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchByTagValue( tag, value );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

bool StarListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{
	    myKids->RemoveCurrent();
	    return true;
	}
    }

    return false;
}

int StarListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of objects inside me:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	curLen = myKids->Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}

List<ASTnode*> * StarFileNode::searchByTag( string  &searchFor )
{
    return myStarList->searchByTag( searchFor );
}

List<ASTnode*> * StarFileNode::searchByTagValue( string &tag, string &value )
{
    return myStarList->searchByTagValue( tag, value );
}

List<ASTnode*> * StarFileNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( myStarList != NULL )
    {
	matches = myStarList->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    return retVal;
}

bool StarFileNode::unlinkChild( ASTnode *child )
{
    return myStarList->unlinkChild( child );
}



StarFileNode::StarFileNode(StarFileNode &n)
{
    (void)copyFrom( n, false );
}

StarFileNode::StarFileNode( bool link, StarFileNode &n)
{
    (void)copyFrom( n, link );
}

bool StarFileNode::copyFrom( StarFileNode &fromThis, bool linkWith )
{

    ASTnode::copyFrom( fromThis );
    myStarList = new StarListNode( linkWith, *(fromThis.myStarList) );
    myStarList->setParent( this );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

StarFileNode::StarFileNode(StarListNode *n) : myStarList(n)
{
    myStarList->setParent( this );
} 

StarFileNode::StarFileNode()
{
  myStarList = new StarListNode;
  myStarList->setParent( this );
}

StarFileNode::~StarFileNode()
{
    amDeleting = true;
    unlinkMe();
    if( myStarList != NULL )
	delete myStarList;
    amDeleting = false;
}

DataNameNode::DataNameNode(const string &str)
    : myStrVal(str)
{}

DataNameNode::DataNameNode(bool link,  DataNameNode &fromThis )
    : myStrVal( fromThis.myStrVal )
{
    if( link )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
}

DataValueNode::DataValueNode(const char *str) :
    myStrVal(str),
    type( NON )
{}
DataValueNode::DataValueNode(const string & str) :
    myStrVal(str),
    type( NON )
{}
DataValueNode::DataValueNode(const string & str, ValType initType) :
    myStrVal(str),
    type( initType )
{
    switch( initType )
    {
	case NON:
	    // do nothing special
	    break;
	case SINGLE:
	    // If there are single-quotes at the start and end, strip them.
	    if( myStrVal[0] == '\'' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\'' )
		myStrVal.erase( (int)(myStrVal.length() - 1), 1 );
	    break;
	case DOUBLE:
	    // If there are double-quotes at the start and end, strip them.
	    if( myStrVal[0] == '\"' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\"' )
		myStrVal.erase( (int)(myStrVal.length() - 1), 1 );
	    break;
	case SEMICOLON:
	    // If there are semicolons at the start or end, strip them.
	    if( myStrVal[0] == ';' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[0] == '\n' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\n' )
		myStrVal.erase( (int)(myStrVal.length() - 1), 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == ';' )
		myStrVal.erase( (int)(myStrVal.length() - 1), 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\n' )
		myStrVal.erase( (int)(myStrVal.length() - 1), 1 );
	    break;
	case FRAMECODE:
	    // If there is a dollar-sign at the start, strip it.
	    if( myStrVal[0] == '$' )
		myStrVal.erase( 0, 1 );
	    break;
    }
}
DataValueNode::DataValueNode(const char *str, ValType initType) :
    myStrVal(str),
    type( initType )
{
    switch( initType )
    {
	case NON:
	    // do nothing special
	    break;
	case SINGLE:
	    // If there are single-quotes at the start and end, strip them.
	    if( myStrVal[0] == '\'' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\'' )
		myStrVal.erase( (int)(myStrVal.length() - 1) , 1 );
	    break;
	case DOUBLE:
	    // If there are double-quotes at the start and end, strip them.
	    if( myStrVal[0] == '\"' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\"' )
		myStrVal.erase( (int)(myStrVal.length() - 1) , 1 );
	    break;
	case SEMICOLON:
	    // If there are semicolons at the start or end, strip them.
	    if( myStrVal[0] == ';' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[0] == '\n' )
		myStrVal.erase( 0, 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == '\n' )
		myStrVal.erase( (int)(myStrVal.length() - 1) , 1 );
	    if( myStrVal[ myStrVal.length() - 1 ] == ';' )
		myStrVal.erase( (int)(myStrVal.length() - 1) , 1 );
	    break;
	case FRAMECODE:
	    // If there is a dollar-sign at the start, strip it.
	    if( myStrVal[0] == '$' )
		myStrVal.erase( 0, 1 );
	    break;
    }
}

DataValueNode::DataValueNode( DataValueNode &copy )
{
    (void)copyFrom( copy, false );
}

DataValueNode::DataValueNode( bool link, DataValueNode &copy )
{
    (void)copyFrom( copy, link );
}

bool DataValueNode::copyFrom( DataValueNode &fromThis, bool linkWith )
{
#ifdef DEBUG_DVN
    cerr << "DEBUG_DVN: copyFrom ptr = " << &fromThis << endl;
    cerr << "DEBUG_DVN: copyFrom val = " << fromThis.myValue() << endl;
#endif DEBUG_DVN
    ASTnode::copyFrom( fromThis );
    myStrVal = fromThis.myName();
    type = fromThis.myDelimType();

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

void DataValueNode::setValue( const string &newVal )
{
    myStrVal = newVal;
}

void DataValueNode::setValue( const char *newVal )
{
    myStrVal = string( newVal );
}

int DataValueNode::myLongestStr( void )
{
    int retVal = myStrVal.length();

    switch( type )
    {
	case NON :        retVal += 1;  // One more for a leading space
			  break;
	case DOUBLE :
	case SINGLE :     retVal += 2;
	                  break;
	case FRAMECODE :  retVal += 1;
	                  break;
	case SEMICOLON :  retVal += 4;  // Probably fairly meaningless, really.
			  break;
	default :         break;  // Makes the compiler shut up.
    }
    return retVal;
}

void DataValueNode::setDelimType( ValType setTo )
{
    type = setTo;
}

bool DataValueNode::operator==( const string &str )
{
    return  (myStrVal == str) ;
}
bool DataValueNode::operator< (const string &str)
{   return  myStrVal < str ;
}
bool DataValueNode::operator> (const string &str)
{   return  myStrVal > str;
}
bool DataValueNode::operator<=(const string &str)
{   return  myStrVal <= str;
}
bool DataValueNode::operator>=(const string &str)
{   return  myStrVal >= str;
}

bool DataValueNode::operator==( const char *str )
{
    // Just a wrapper around the other version of this function.
    // ---------------------------------------------------------
    bool     retVal;
    string   *tempStr = new string( str );
    retVal = ( operator==(*tempStr) );
    delete tempStr;
    return retVal;
}

bool DataValueNode::operator< (const char *str)
{   return  myStrVal < string(str) ;
}
bool DataValueNode::operator> (const char *str)
{   return  myStrVal > string(str);
}
bool DataValueNode::operator<=(const char *str)
{   return  myStrVal <= string(str);
}
bool DataValueNode::operator>=(const char *str)
{   return  myStrVal >= string(str);
}

List<ASTnode*> * DataValueNode::searchForType(
	     ASTtype    type,
	     int        delim
	     )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;

    if( delim < 0 || delim == myDelimType() )  /* compare int to enum is legal, but be careful */
	if( isOfType( type ) )
	    retVal->AddToEnd( this );
    return retVal;
}

DataItemNode::DataItemNode(DataNameNode *name, DataValueNode *value) 
:
 myDataName(name), myDataValue(value)
{
    myDataName->setParent(this);
    myDataValue->setParent(this);
}

///////////////////////////////////////////////////////////////////////////
//Copy constructor for DataItemNode
//////////////////////////////////////////////////////////////////////////
DataItemNode::DataItemNode( DataItemNode &  D)
{
    (void)copyFrom( D, false );
}

DataItemNode::DataItemNode( bool link, DataItemNode & D)
{
    (void)copyFrom( D, link );
}

bool DataItemNode::copyFrom( DataItemNode &fromThis, bool linkWith )
{
    ASTnode::copyFrom( fromThis );
    myDataName = new DataNameNode( linkWith, *(fromThis.myDataName) );
    myDataValue = new DataValueNode( linkWith, *(fromThis.myDataValue) );

    myDataName->setParent(this);
    myDataValue->setParent(this);

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}
/////////////////////////////////////////////////////////////////////////

DataItemNode::DataItemNode(const string & name, const string & value)
{
  myDataName = new DataNameNode(name);

  DataValueNode *newValue = new DataValueNode( value );

  myDataValue = newValue;

  myDataName->setParent(this);
  myDataValue->setParent(this);
}

DataItemNode::DataItemNode( const string & name,
	                    const string & value,
			    DataValueNode::ValType  type  )
{
  myDataName = new DataNameNode(name);

  DataValueNode *newValue = new DataValueNode( value, type );

  myDataValue = newValue;

  myDataName->setParent(this);
  myDataValue->setParent(this);
}

DataItemNode::~DataItemNode()
{
    amDeleting = true;
    unlinkMe();
    if( myDataName != NULL )
	delete myDataName;
    if( myDataValue != NULL )
	delete myDataValue;
    amDeleting = false;
}

void DataItemNode::setDelimType( DataValueNode::ValType newType )
{
    if( myDataValue != (DataValueNode*)NULL )
        myDataValue->setDelimType( newType );
}

void DataItemNode::setValue( const string &val )
{
    if( myDataValue != (DataValueNode*)NULL )
	myDataValue->setValue( val );
}

void DataItemNode::setValue( const char  *val )
{
    if( myDataValue != (DataValueNode*)NULL )
	myDataValue->setValue( val );
}
void DataItemNode::setName( const string &name )
{
    if( myDataName != (DataNameNode*)NULL )
	myDataName->setName( name );
}

void DataItemNode::setName( const char   *name )
{
    if( myDataName != (DataNameNode*)NULL )
	myDataName->setName( name );
}


List<ASTnode*> * DataItemNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;

    if( myDataName != NULL )
	if( myDataName->myName() == searchFor )
	    retVal->AddToEnd( this );

    return retVal;
}

List<ASTnode*> * DataItemNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( myDataName != NULL )
    {
	matches = myDataName->searchForType( type, delim  );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    if( myDataValue != NULL )
    {
	matches = myDataValue->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    return retVal;
}


bool DataItemNode::unlinkChild( ASTnode *child )
{
    if( myDataName == (DataNameNode*)child )
    {   myDataName = (DataNameNode*)NULL;
        return true;
    }
    else if( myDataValue == (DataValueNode*)child )
    {   myDataValue = (DataValueNode*)NULL;
	return true;
    }
    return false;
}

List<ASTnode*> * DataItemNode::searchByTagValue( string &tag,
	                                  string &value )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;

    if( myDataName != NULL )
	if( myDataName->myName() == tag )
	    if( *myDataValue == value )
		retVal->AddToEnd( this );

    return  retVal;
}


LoopNameListNode::LoopNameListNode(ASTlist<DataNameNode *> *L)
  :
  myKids(L) 
{
    myKids->setParent(this);
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	myKids->Current()->setParent(this);
}

LoopNameListNode::LoopNameListNode( vector<DataNameNode *> &L)
{
    vector<DataNameNode *>::iterator  i;

    myKids = new ASTlist<DataNameNode *>;
    myKids->setParent(this);

    for( i = L.begin() ; i != L.end() ; i++ )
        myKids->AddToEnd( new DataNameNode( **i ) );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        myKids->Current()->setParent(this);

}

LoopNameListNode::LoopNameListNode( LoopDefListNode  &L )
{
    (void)copyFrom( L, false );
}

LoopNameListNode::LoopNameListNode( bool link, LoopDefListNode  &L )
{
    (void)copyFrom( L, link );
}

bool LoopNameListNode::copyFrom( LoopDefListNode &L, bool linkWith )
{
    DataNameNode *remember = (L.myKids->AtEnd()) ? 
			(DataNameNode*)NULL : L.myKids->Current();
    ASTnode::copyFrom( L );
    myKids = new ASTlist<DataNameNode *>;
    myKids->setParent(this);

    for( L.myKids->Reset() ; ! L.myKids->AtEnd() ; L.myKids->Next() )
        myKids->AddToEnd( new DataNameNode( linkWith, *(L.myKids->Current()) )  );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        myKids->Current()->setParent(this);

    (void)L.myKids->SeekTo( remember );

    if( linkWith )
    {   setPeer( &L );
	L.setPeer( this );
    }
    return true;
}

LoopNameListNode::LoopNameListNode()
{
  myKids = new ASTlist<DataNameNode *>;
  myKids->setParent(this);
  for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
      myKids->Current()->setParent(this);
}

LoopNameListNode::~LoopNameListNode()
{
    amDeleting = true;
    unlinkMe();
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	if( myKids->Current() != NULL )
	    delete myKids->Current();

    if( myKids != NULL )
	delete myKids;
    amDeleting = false;
}

int LoopNameListNode::isInLoop( void )
{
    bool    foundLoop = false;
    int     count = 0;
    ASTnode  *ancestor;

    for(   ancestor = this ;
	   ancestor != NULL && foundLoop == false  ;
	   ancestor = ancestor->myParent() )
    {
	if( ancestor->isOfType( ASTnode::DATALOOPNODE ) )
	{
	    foundLoop = true;
	    break;
	}
	if( ancestor->isOfType( ASTnode::DATALOOPNAMELISTNODE ) )
	{
	    // Find out where in the major list this minor list is:
	    DataLoopNameListNode::iterator  i;
	    DataLoopNameListNode            *parent =
				(DataLoopNameListNode*)ancestor;
	    for(    i = parent->begin(), count = 1 ;
		    i != parent->end() && *i != this ;
		    i++, count++ )
	    { }
	}
    }
    if( ancestor == NULL )
	return 0;
    else
	return count;
}

void LoopNameListNode::insert( iterator pos, DataNameNode &v )
{
    insert( pos, &v, DataValueNode( string(".") ) );
}

void LoopNameListNode::insert( iterator pos, DataNameNode *v )
{
    insert( pos, v, DataValueNode( string(".") ) );
}

void LoopNameListNode::insert(
	iterator       pos,
	DataNameNode   &v,
	DataValueNode  &val )
{
    DataNameNode   *newNode = new DataNameNode( v );
    insert( pos, newNode, val );
}

void LoopNameListNode::insert(
	iterator       pos,
	DataNameNode   *v,
	DataValueNode  &val )
{
    iterator       i;
    int            rowPos, nest;
    ASTnode        *ancestor;
    DataLoopNode  *ancestorLoop;
    LoopNode       *curLoop;


    // Need to count: What position is this new tag in the row?
    for(    i = begin() , rowPos = 0 ;
	    i != end() && i != pos;
	    i++, rowPos++ )
    { }

    // Need to find out if I am in a loop, and how many levels
    // nested down I am:
    nest = isInLoop();

    // If I am in a loop, then then I need to iterate over every
    // row at my nesting level and insert the new value into them:
    if( nest )
    {
	// Get the loop I am in (Guaranteed to be there because I already
	// checked via isInLoop() ):
	for(    ancestor = this ;
		! ancestor->isOfType( ASTnode::DATALOOPNODE );
		ancestor = ancestor->myParent() )
	{ }
	ancestorLoop = (DataLoopNode*)ancestor;

	(ancestorLoop->getVals()).insertAtPos( nest-1, rowPos, val );
    }

    myKids->insert( &*pos, v );
}

void LoopNameListNode::erase( iterator pos )
{
    iterator       i;
    int            rowPos, nest;
    ASTnode        *ancestor;
    DataLoopNode  *ancestorLoop;
    LoopNode       *curLoop;


    // Need to count: What position is this new tag in the row?
    for(    i = begin() , rowPos = 0 ;
	    i != end() && i != pos;
	    i++, rowPos++ )
    { }

    // Need to find out if I am in a loop, and how many levels
    // nested down I am:
    nest = isInLoop();

    // If I am in a loop, then then I need to iterate over every
    // row at my nesting level and erase the value:
    if( nest )
    {

	// Get the loop I am in (Guaranteed to be there because I already
	// checked via isInLoop() ):
	for(    ancestor = this ;
		! ancestor->isOfType(ASTnode::DATALOOPNODE);
		ancestor = ancestor->myParent() )
	{ }
	ancestorLoop = (DataLoopNode*)ancestor;

	ancestorLoop->getVals().eraseAtPos( nest-1, rowPos );
	
    }

    delete *pos;
    myKids->erase( &*pos ) ;

    // If I am now gone entirely, then just get rid of this
    // nesting level altogether and all nesting levels under it:
    if( size() == 0 )
	if( nest )
	    ancestorLoop->getNames().erase(
		    &( ancestorLoop->getNames()[nest-1] ) );
}

void LoopNameListNode::erase( iterator from, iterator to )
{
    iterator       i;
    DataNameNode   *namePtr;
    int            fromNum, toNum, rowPos;
    ASTnode        *ancestor;
    DataValueNode  *ancestorLoop;
    LoopNode       *curLoop;


    // Need to count: What position is the from and to tag in the row?
    for(    i = begin() , rowPos = 0 ;
	    i != end() ;
	    i++, rowPos++ )
    {
	if( i == from )
	    fromNum = rowPos;
	if( i == to )
	    toNum = rowPos;
    }

    // erase the values one at a time using the above function.
    while( toNum != fromNum )
    {
	namePtr = (*myKids)[fromNum];
	delete namePtr;
        erase( &namePtr );
	toNum--;
    }
}

void LoopTableNode::insertAtPos(
	int            nest,
	int            rowPos,
	DataValueNode  &val )

{
    iterator row;

    // This algorithm is recursive: If this loop is above the
    // nesting level we want then iterate over the rows looking
    // for sub-loops and pass the recursive baton to them.  If
    // this loop *is* the nesting level we want, then insert the
    // value to each row of the loop:

    // Base Case:
    if( nest == 0 )
	for( row = begin() ; row != end() ; row++ )
	{
	    // NOTE: by calling myVals.insert() instead of just insert(),
	    // I avoid the shoot-self-in-foot protections that are in
	    // the public insert() function.  (They would never let me add
	    // a value to a row in a loop).  I can do this because the
	    // LoopRowNode class considers me its friend.
	    DataValueNode *newNode = new DataValueNode( val ) ;
	    row->myVals.insert( &((**row)[ rowPos ]), newNode);
	}
    // Recursive Case:
    else
	for( row = begin() ; row != end() ; row++ )
	    if( row->myLoop() )
		row->myLoop()->insertAtPos( nest - 1, rowPos, val );
}

void LoopTableNode::eraseAtPos(
	int                  nest,
	int                  rowPos )

{
    iterator row;

    // This algorithm is recursive: If this loop is above the
    // nesting level we want then iterate over the rows looking
    // for sub-loops and pass the recursive baton to them.  If
    // this loop *is* the nesting level we want, then erase the
    // value to each row of the loop:

    // Base Case:
    if( nest == 0 )
	for( row = begin() ; row != end() ; row++ )
	{
	    // NOTE: by calling myVals.erase() instead of just erase(),
	    // I avoid the shoot-self-in-foot protections that are in
	    // the public erase() function.  (They would never let me
	    // remove a value in an existing loop).  I can do this
	    // because the LoopRowNode class considers me its friend.
	    row->myVals.erase( &((**row)[ rowPos ]) );

	}
    // Recursive Case:
    else
	for( row = begin() ; row != end() ; row++ )
	    if( row->myLoop() )
		row->myLoop()->eraseAtPos( nest - 1, rowPos );
}

void LoopNameListNode::reset()
{
  myKids->Reset();
}

List<ASTnode*> * LoopNameListNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	if( myKids->Current()->myName() == searchFor )
	    retVal->AddToEnd( myKids->Current() );

    return retVal;
}

List<ASTnode*> * LoopNameListNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );


    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	matches = myKids->Current()->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    return retVal;
}


bool LoopNameListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}


void LoopNameListNode::getNextTag()
{
  if (myKids->AtEnd())
    myKids->Reset();

  myKids->Next();
}

bool LoopNameListNode::atEnd()
{
  return myKids->AtEnd();
}

bool LoopNameListNode::atStart()
{
  return myKids->AtStart();
}

LoopValListNode::LoopValListNode(ASTlist<DataValueNode *> *L)
  :
  myKids(L) 
{
    myKids->setParent(this);
}

LoopValListNode::LoopValListNode()
{
  myKids = new ASTlist<DataValueNode *>;
  myKids->setParent(this);
}

LoopValListNode::~LoopValListNode()
{
    amDeleting = true;
  unlinkMe();
  if( myKids != NULL )
      delete myKids;       // The individual DataValueNodes are not deleted because
                           // now they are part of the transformed structure
    amDeleting = false;
}


DataLoopValListNode::Status LoopValListNode::getNextValue(DataValueNode* & value)
{
  if (myKids->AtEnd())
    return DataLoopValListNode::STOP;

  value = myKids->Current();
  myKids->Next();
  return DataLoopValListNode::CONTINUE;
}

void DataLoopValListNode::debugPrint( void )
{
    ASTlist<LoopValListNode*>::iterator  i;

    for( i = myKids->begin() ; i != myKids->end() ; i++ )
    {
	fprintf( stderr, "\tA valList:\n" );
	(*i)->debugPrint();
    }
}

void LoopValListNode::debugPrint( void )
{
    ASTlist<DataValueNode*>::iterator  i;

    for( i = myKids->begin() ; i != myKids->end() ; i++ )
    {
	fprintf( stderr, "\t\t%s\n", (*i)->myValue().c_str() );
    }
}


void LoopValListNode::reset()
{
  myKids->Reset();
}

bool DataLoopValListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

bool LoopValListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

int LoopValListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of values in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	curLen = myKids->Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}

DataLoopNameListNode::DataLoopNameListNode(ASTlist<LoopDefListNode *> *L) 
  : 
  myKids(L) 
{
  myKids->setParent(this);
  for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
      myKids->Current()->setParent(this);
}

DataLoopNameListNode::DataLoopNameListNode( DataLoopDefListNode &L)
{
    (void)copyFrom( L, false );
}

DataLoopNameListNode::DataLoopNameListNode( bool link, DataLoopDefListNode &L)
{
    (void)copyFrom( L, link );
}

bool DataLoopNameListNode::copyFrom( DataLoopDefListNode &fromThis, bool linkWith )
{
    LoopNameListNode *remember = (fromThis.myKids->AtEnd()) ? 
			(LoopNameListNode*)NULL : fromThis.myKids->Current();
    ASTnode::copyFrom( fromThis );
    myKids = new ASTlist<LoopNameListNode *>;
    myKids->setParent(this);

    for( fromThis.myKids->Reset() ; ! fromThis.myKids->AtEnd() ; fromThis.myKids->Next() )
        myKids->AddToEnd( new LoopNameListNode( linkWith, *(fromThis.myKids->Current()) )  );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        myKids->Current()->setParent(this);

    (void)fromThis.myKids->SeekTo( remember );
    currRow = fromThis.currRow;

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

DataLoopNameListNode::DataLoopNameListNode()
{
  myKids = new ASTlist<LoopNameListNode *>;
  myKids->setParent(this);
  for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
      myKids->Current()->setParent(this);
}


DataLoopNameListNode::~DataLoopNameListNode()
{
    amDeleting = true;
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	if( myKids->Current() != NULL )
	    delete myKids->Current();

    if( myKids != NULL )
	delete myKids;
    unlinkMe();
    amDeleting = false;
}

void DataLoopNameListNode::debugPrint( void )
{
    ASTlist<LoopNameListNode*>::iterator i;

    for( i = myKids->begin() ; i != myKids->end() ; i++ )
    {
	fprintf( stderr, "\tList of names:\n" );
	(*i)->debugPrint();
    }
}

void LoopNameListNode::debugPrint( void )
{
    ASTlist<DataNameNode*>::iterator i;

    for( i = myKids->begin() ; i != myKids->end() ; i++ )
    {
	fprintf( stderr, "\t\t%s\n", (*i)->myName().c_str() );
    }
}

bool DataLoopNameListNode::insert(
	iterator               pos,
	vector<DataNameNode*>  &v )
{
    // Let the other overloaded version do all the work:
    LoopNameListNode *newNode = new LoopNameListNode( v );
    return insert( pos, newNode );
}

bool DataLoopNameListNode::insert(
	iterator          pos,
	LoopNameListNode  &v  )
{
    // Let the other overloaded version do all the work:
    return insert( pos, new LoopNameListNode( v ) );
}

bool DataLoopNameListNode::insert(
	iterator          pos,
	LoopNameListNode  *v  )
{
    ASTnode   *myParentLoop;

    // Need to see if we are in DataLoopNode:
    // --------------------------------------
    for(    myParentLoop = this
	    ;
	    myParentLoop != NULL &&
	    !(myParentLoop->isOfType(ASTnode::DATALOOPNODE))
	    ;
	    myParentLoop = myParentLoop->myParent()
	)
    { }
    // If we are in a DataLoopNode:
    // ----------------------------
    if( myParentLoop->isOfType(ASTnode::DATALOOPNODE) )
    {
	// If this is not the deepest nesting level, then
	// disallow the insertion:
	if( pos != end() )
	    return false;
	// Ask the associated DataLoopNode to make a new stub
	// nest level:
	( ((DataLoopNode*)myParentLoop)->getVals() ).makeNewNestLevel();
    }
    myKids->insert( &*pos, v );
    return true;
}

void DataLoopNameListNode::erase( iterator pos )
{
    DataLoopNameListNode::iterator  i;
    LoopNameListNode::iterator      oneName;
    int                             thisDepth;
    int                             curDepth;
    ASTnode                         *ancestorLoop;

    for(    ancestorLoop = this ;
	    ancestorLoop != NULL && ! ancestorLoop->isOfType( ASTnode::DATALOOPNODE ) ;
	    ancestorLoop = ancestorLoop->myParent()   )
    { }

    // Note, because of the rules of STAR files, I cannot
    // leave an inner loop under a loop level that no longer
    // exists.  Therefore when I delete a row I must delete all
    // the ones after it as well.

    // So I start at the deepest level and begin deleting upward to
    // the current level.  First, find the index number of the
    // position passed.  This is vital because we have to refer to
    // everything by index number not iterator, since the iterators
    // can become invalid every time the vector gets realloced, which
    // could very easily happen when we are deleting things from it:
    for(    i = begin(), thisDepth = 0 ;
	    i != end() && i != pos ;
	    i++, thisDepth++ )
    { }
    assert( i != end() ); // If this fails the caller tried to delete
                          // something that doesn't exist.

    // Now iterate over the list of names backward, so as to delete
    // things depth-first:
    for(    curDepth = size() - 1 ;
	    curDepth >= thisDepth ;
	    curDepth--   )
    {
	// Delete the names one at a time, thereby activating
	// the code I've already written that iterates through
	// and removes the associated columns.  If this is too
	// slow, this can be reimplemented more complicatedly in
	// the future without hurting the interface.
	for(    oneName = (*this)[curDepth]->begin() ;
		oneName != (*this)[curDepth]->end() ;
		oneName++     )
	    (*this)[curDepth]->erase( oneName );

	delete (*this)[curDepth];
    }
    if( ancestorLoop != NULL )
    {
	((DataLoopNode*)ancestorLoop)->getValsPtr()->truncateNestLevel( thisDepth );
    }
}

DataLoopNameListNode::Status DataLoopNameListNode::getNextTag(
					     DataLoopValListNode::Status vstat,
					     int & row)
{
    Status stat;
    bool   popUpLevel = false;

    if (myKids->AtEnd())
       return ERROR;

    popUpLevel = false; // default.

    // First, if this is a normal value, then we already determined the right
    // nest level (confusingly called "currRow") the last time this was ran:
    if( vstat == DataLoopValListNode::CONTINUE )
    {
	row = currRow;
	if( myKids->Current()->atStart() )
	    stat = NEW_ITER; // Just beginning a new row of a loop.
	else
	    stat = SAME_ITER; // Same row as last time.
	
	// Increment to the next tagName to get ready for the next
	// call to this function:
	myKids->Current()->getNextTag();

	// If we just went past the end of the names for this nestlevel,
	// then rewind it and drop to the next nest level:
	if( myKids->Current()->atEnd() )
	{
	    myKids->Current()->reset();

	    myKids->Next();
	    currRow++;

	    // But wait, maybe that was the innermost nest level, in
	    // which case we shouldn't have gone any further down:
	    if( myKids->AtEnd() )
		popUpLevel = true;
	}
    }
    // Else, if it is a stop value, then the current nesting level
    // is done and we pop back one nest level:
    else if( vstat == DataLoopValListNode::STOP )
    {
	row = currRow - 1;
	popUpLevel = true;
	stat = STOP;
    }
    else
	stat = ERROR;

    if( popUpLevel )
    {
	// Either a STOP was encountered or the last value of a row
	// at the innermost nesting level was encountered:  Either
	// way we go up a nest level:
	myKids->Prev();
	currRow--;
    }
    
    return stat;
}

void DataLoopNameListNode::reset()
{
  currRow = 0;

  for (myKids->Reset(); !myKids->AtEnd(); myKids->Next())
    myKids->Current()->reset();

  myKids->Reset();
}

List<ASTnode*> * DataLoopNameListNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchByTag( searchFor );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * DataLoopNameListNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

bool DataLoopNameListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

int DataLoopNameListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	curLen = myKids->Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}


int LoopNameListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	curLen = myKids->Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}


DataLoopValListNode::DataLoopValListNode(ASTlist<LoopValListNode *> *L) 
: 
  myKids(L)
{
    myKids->setParent(this);
}

DataLoopValListNode::~DataLoopValListNode()
{
  amDeleting = true;
  unlinkMe();
  for (myKids->Reset(); !myKids->AtEnd(); myKids->Next())
    if( myKids->Current() != NULL )
	delete myKids->Current();

  if( myKids != NULL )
      delete myKids;
  amDeleting = false;
}

DataLoopValListNode::Status DataLoopValListNode::getNextValue(DataValueNode* &value)
{
  if (myKids->AtEnd())
    return END;

  Status stat = myKids->Current()->getNextValue(value);

  if (stat == STOP) {
    myKids->Next();

    if (!myKids->AtEnd())
      myKids->Current()->reset();

    return STOP;
  }
  else
    return CONTINUE;
}

void DataLoopValListNode::reset()
{
  myKids->Reset();
  if (!myKids->AtEnd())
    myKids->Current()->reset();
}

int DataLoopValListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	curLen = myKids->Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}

LoopRowNode::IterNode(const bool tflag) : myTflag(tflag)
{
  innerLoop = NULL;
  myVals.setParent(this);
  for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
      myVals.Current()->setParent(this);
}

//////////////////////////////////////////////////////////////////
//Copy constructor for LoopRowNode
/////////////////////////////////////////////////////////////////
LoopRowNode::LoopRowNode(IterNode& N)
{
    (void)copyFrom( N, false );
}

LoopRowNode::LoopRowNode( bool link, IterNode& N)
{
    (void)copyFrom( N, link );
}

LoopRowNode::LoopRowNode( vector<DataValueNode*> &v, const bool tflag )
    : myTflag(tflag)
{
    vector<DataValueNode*>::iterator  i;

    innerLoop = NULL;
    myVals.setParent(this);

    // Frivolous: myVals should be empty at this point, since this is
    // a newly constructed object:
    for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
         myVals.Current()->setParent(this);

    for( i = v.begin() ; i != v.end() ; i++ )
	myVals.insert( myVals.end(), new DataValueNode( **i ) );
}

bool LoopRowNode::copyFrom( IterNode &fromThis, bool linkWith )
{
    LoopRowNode &N = fromThis;

    ASTnode::copyFrom( fromThis );
    myTflag = N.myTflag;
    DataValueNode *remember = (N.myVals.AtEnd()) ? 
			(DataValueNode*)NULL : N.myVals.Current();
    innerLoop = ( N.innerLoop == NULL ) ?
		(LoopTableNode*)NULL : new LoopIter( linkWith, *(N.innerLoop) );

    myVals.FreeList();
    myVals.setParent(this);
    for( N.myVals.Reset() ; ! N.myVals.AtEnd() ; N.myVals.Next() )
    {
        myVals.AddToEnd( new DataValueNode( linkWith, *(N.myVals.Current()) ) );
    }
    for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
        myVals.Current()->setParent(this);
    if( innerLoop != NULL )
        innerLoop->setParent(this);
    (void)N.myVals.SeekTo( remember );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}
///////////////////////////////////////////////////////////////////

LoopRowNode::~LoopRowNode()
{
  amDeleting = true;
  for (myVals.Reset(); !myVals.AtEnd(); myVals.Next())
    if( myVals.Current() != NULL )
	delete myVals.Current();

  if( innerLoop != NULL )
      delete innerLoop;
  unlinkMe();
  amDeleting = false;
}


bool LoopRowNode::insert( iterator pos, DataValueNode &v )
{
    return insert( pos, new DataValueNode( v ) );
}

bool LoopRowNode::insert( iterator pos, DataValueNode *v )
{
    if( isInLoop() )
	return false;

    // This LoopRowNode is not attached to a loop yet, so go ahead and insert:

    myVals.insert( &*pos, v );
    return true;
}

bool LoopRowNode::erase( iterator pos )
{
    if( isInLoop() )
	return false;


    delete *pos;  // Deletes the DataVaueNode pointed to.
    myVals.erase( &*pos );
    return true;
}

bool LoopRowNode::erase( iterator from, iterator to )
{
    iterator  i;

    if( isInLoop() )
	return false;

    for( i = from ; i != to ; i++ )
    {
	delete *i;  // Deletes the DataValueNode pointed to.
	myVals.erase( &*i );
    }
    return true;
}

int LoopRowNode::isInLoop( void )
{
    bool    foundLoop = false;
    int     count = 0;
    ASTnode  *ancestor;

    for(   ancestor = this ;
	   ancestor != NULL && foundLoop == false  ;
	   ancestor = ancestor->myParent() )
    {
	if( ancestor->isOfType( ASTnode::DATALOOPNODE ) )
	    foundLoop = true;
	if( ancestor->isOfType( ASTnode::LOOPTABLENODE ) )
	    count++;  // Every LoopNode encountered on the way up
	              // the parent is another level of nesting.
    }
    if( ancestor == NULL )
	return 0;
    else
	return count;
}

void LoopRowNode::addValue(DataValueNode *value)
{
  value->setParent(this);
  myVals.AddToEnd(value);
}

void LoopRowNode::addInnerLoopValue(const DataLoopNameListNode::Status dstat,
				 const int row,
				 const int currRow,
				 DataValueNode *value)
{
  if (!innerLoop)
    if (myTflag)
      innerLoop = new LoopTableNode(string("tabulate"));
    else
      innerLoop = new LoopTableNode(string("linear"));

  innerLoop->setParent(this);

  if( dstat != DataLoopNameListNode::STOP )
      innerLoop->addValue(dstat, row, currRow, value);
}

bool LoopRowNode::unlinkChild( ASTnode *child )
{
    if( innerLoop == child )
    {   innerLoop = (LoopTableNode*)NULL;
	return true;
    }
    for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
    {   if( myVals.Current() == child )
	{   myVals.RemoveCurrent();
	    return true;
 	}
    }
    return false; // pessamistic default.
}


List <ASTnode*> *LoopRowNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( innerLoop != NULL )
    {
	matches = innerLoop->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
    {
	matches = myVals.Current()->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    return retVal;
}

int LoopRowNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myVals.Reset() ; ! myVals.AtEnd() ; myVals.Next() )
    {
	curLen = myVals.Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}


///////////////////////////////////////////////////////////////////////////
//Copy constructor for LoopTableNode Node
//////////////////////////////////////////////////////////////////////////

LoopTableNode::LoopIter(LoopIter& I)
{
    (void)copyFrom( I, false );
}

LoopTableNode::LoopIter( bool link, LoopIter& I)
{
    (void)copyFrom( I, link );
}

bool LoopTableNode::copyFrom( LoopIter &fromThis, bool linkWith )
{
    ASTnode::copyFrom( fromThis );
    myTflag = fromThis.myTflag;
    indentFlag = fromThis.indentFlag;
    rowsPerLine = fromThis.rowsPerLine;

    LoopRowNode *remember = (fromThis.myIters.AtEnd()) ? 
				(LoopRowNode*)NULL : fromThis.myIters.Current();
    myIters.FreeList();
    myIters.setParent(this);
    for( fromThis.myIters.Reset() ; ! fromThis.myIters.AtEnd() ; fromThis.myIters.Next() )
	myIters.AddToEnd( new LoopRowNode( linkWith, *(fromThis.myIters.Current()) )   );
    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
	myIters.Current()->setParent(this);
    (void)fromThis.myIters.SeekTo( remember );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}
//////////////////////////////////////////////////////////////////////////


LoopTableNode::LoopTableNode(const string & tflag, bool indentFl, int rowsPerLn )
{
  myIters.setParent(this);
  if (tflag == "tabulate")
    myTflag = true;
  else
    myTflag = false;
  indentFlag = indentFl;
  rowsPerLine = rowsPerLn;
}

LoopTableNode::LoopTableNode(bool tflag )
{
  myIters.setParent(this);
  myTflag = tflag;
  indentFlag = true;
  rowsPerLine = 1;
}

LoopTableNode::~LoopTableNode()
{
  amDeleting = true;
  unlinkMe();
  for (myIters.Reset(); !myIters.AtEnd(); myIters.Next())
    if( myIters.Current() != NULL )
	delete myIters.Current();
  amDeleting = false;
}

bool LoopTableNode::insert(
	iterator    pos,
	vector<DataValueNode*> &v,
	const bool  tflag )
{
    LoopRowNode  *newRow;
    if( v.size() != getNumCols() )
    {
	return false;
    }

    newRow = new LoopRowNode( v , tflag );
    myIters.insert( &*pos, newRow );

    // If I am not at the deepest level and I have no
    // inner loop, make an empty inner loop so that there
    // will be a "stop_" printed:
    if( newRow->innerLoop == NULL )
    {
	if( getMyDepth() < getMaxDepth() )
	{
	    newRow->innerLoop = new LoopTableNode( tflag );
	    newRow->innerLoop->setParent( newRow );
	}
    }
    return true;
}

bool LoopTableNode::insert( iterator pos, LoopRowNode &v )
{
    return insert( pos, new LoopRowNode( v ) );
}

bool LoopTableNode::insert( iterator pos, LoopRowNode *v )
{
    if( v->size() != getNumCols() )
    {
	return false;
    }

    myIters.insert( &*pos, v );

    // If I am not at the deepest level and I have no
    // inner loop, make an empty inner loop so that there
    // will be a "stop_" printed:
    if( v->innerLoop == NULL )
    {
	if( getMyDepth() < getMaxDepth() )
	{
	    v->innerLoop = new LoopTableNode( v->myTflag );
	    v->innerLoop->setParent( v );
	}
    }
    return true;
}

bool LoopTableNode::insert( iterator pos, LoopTableNode &t )
{
    return insert( pos, new LoopTableNode( t ) );
}

bool LoopTableNode::insert( iterator pos, LoopTableNode *t )
{
    LoopRowNode  *newRow;
    iterator     i;

    for( i = t->begin() ; i != t->end() ; i++ )
    {
	if( (*i)->size() != getNumCols() )
	    return false;
	newRow = new LoopRowNode( **i );
        myIters.insert( &*pos, newRow );

	// If I am not at the deepest level and I have no
	// inner loop, make an empty inner loop so that there
	// will be a "stop_" printed:
	if( newRow->innerLoop == NULL )
	{
	    if( getMyDepth() < getMaxDepth() )
	    {
		newRow->innerLoop = new LoopTableNode( newRow->myTflag );
		newRow->innerLoop->setParent( newRow );
	    }
	}
    }
    return true;
}

void LoopTableNode::erase( iterator pos )
{
    iterator i;

    delete *pos;
    myIters.erase( &*pos );

}

void LoopTableNode::erase( iterator from, iterator to )
{
    iterator i;

    // delete the things being pointed at too:
    for( i = from ; i != to ; i++ )
    {
	delete *i;  // deletes the LoopRowNode, not the pointer to it.
	myIters.erase( &*i );
    }
}

void LoopTableNode::truncateNestLevel( int depth )
{
    iterator   i;

    // If it's talking about me: then commit suicide:
    if( depth == 0 )
    {
	delete this;
    }
    // Otherwise recursively drop a level:
    else
    {
	for( i = begin() ; i != end() ; i++ )
	{
	    if( (*i)->myLoop() != NULL )
		(*i)->myLoop()->truncateNestLevel( depth - 1 );
	}
    }
}

int LoopTableNode::getNumCols( void )
{
    ASTnode  *ancestor;
    int      retVal = 0;
    int      nestLevel;

    // Get the DataLoopNode that I am inside of:
    for(   ancestor = this , nestLevel = -1;
	   ancestor != NULL  && ! ancestor->isOfType( ASTnode::DATALOOPNODE ) ;
	   ancestor = ancestor->myParent()  )
    { 
	if( ancestor->isOfType( ASTnode::LOOPTABLENODE ) )
	    nestLevel++;
    }
    if( ancestor == NULL )  // Not contained inside a DataLoopNode
	return 0;
    if( nestLevel < 0 ) //  weird should never happen case.
	return 0;

    // TODO: Need to implement a ton of stuff in the datanamelist class
    // to pull off the following stunt:
    return  (((DataLoopNode*)ancestor)->getNames())[nestLevel]->size() ;
}

int LoopTableNode::getMyDepth( void )
{
    ASTnode *ancestor;
    int     retVal;

    for(    ancestor = this , retVal = -1;
	    ancestor != NULL && ! ancestor->isOfType(ASTnode::DATALOOPNODE) ;
	    ancestor = ancestor->myParent()  )
    {
	if(   ancestor->isOfType(ASTnode::LOOPTABLENODE)  )
	    retVal++;
    }

    return retVal;
}

void LoopTableNode::makeNewNestLevel( void )
{
    iterator i;

    for( i = begin() ; i != end() ; i++ )
	if( (*i)->innerLoop == NULL )
	{
	    (*i)->innerLoop = new LoopTableNode( myTflag );
	    (*i)->innerLoop->setParent( (*i) );
	}
	else
	    (*i)->innerLoop->makeNewNestLevel();
}

int LoopTableNode::getMaxDepth( void )
{
    ASTnode *ancestor;
    int     retVal;

    for(    ancestor = this ;
	    ancestor != NULL && ! ancestor->isOfType(ASTnode::DATALOOPNODE) ;
	    ancestor = ancestor->myParent()  )
    { }
    if( ancestor->isOfType(ASTnode::DATALOOPNODE) )
	retVal = ((DataLoopNode*)ancestor)->getMaxDepth();
    else
	retVal = -1;

    return retVal;
}

void LoopTableNode::addValue(const DataLoopNameListNode::Status dstat,
			const int row,
			const int currRow,
			DataValueNode *value)
{
  LoopRowNode *lastIter = myIters.Last();

  if( row == currRow && dstat != DataLoopNameListNode::STOP )
    if (dstat == DataLoopNameListNode::SAME_ITER)
      lastIter->addValue(value);
    else {
      lastIter = new LoopRowNode(myTflag);
      lastIter->setParent(this);
      lastIter->addValue(value);
      myIters.AddToEnd(lastIter);
    }
  else
  {
    lastIter->addInnerLoopValue(dstat, row, currRow + 1, value);
  }
}

bool LoopTableNode::unlinkChild( ASTnode *child )
{
    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
    {   if( myIters.Current() == child )
	{   myIters.RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

List<ASTnode*> * LoopTableNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    // WARNING:   AN UGLY HACK FOR SPEED PURPOSES HERE:
    //     Over 50% of the data in a typical STAR file is
    //     contained inside loops.  To sift down through
    //     all the values in the loop when we already know
    //     it is impossible to find a match of the kind we
    //     are looking for is a huge time waster.  So here,
    //     we truncate the search at this point if we know that
    //     the type being searched for is one that cannot exist
    //     inside a loop.  I consider this an ugly hack because
    //     it requires that this code has knowlege of how the
    //     entire tree is laid out.  This means the code is
    //     built on the fragile assumption that the tree structure
    //     used in this library won't change.  Theoretically,
    //     this search-truncation could have been done at
    //     every level of the tree for even more speed, but then
    //     it would have been a lot of fragile assumptions built
    //     on top of fragile assumptions.
    //
    switch( type )
    {
	case    ASTNODE :
	case    DATAVALUENODE :
	case	LOOPROWNODE :
	case	LOOPTABLENODE :

		    if( isOfType( type ) )
			retVal->AddToEnd( this );

		    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
		    {
			matches = myIters.Current()->searchForType( type, delim );
			if( matches->Length() > 0 )
			    retVal->AddToEnd( *matches );
			delete matches;
		    }

		    break;
	default:
		    {} /* If it isn't the right type, Don't waste time iterating
			  through the loop - just truncate the search here. */
    }
    return retVal;
}

int LoopTableNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
    {
	curLen = myIters.Current()->myLongestStr();
	if( curLen > retVal )
	    retVal = curLen;
    }
    return retVal;
}

void LoopTableNode::calcPrintSizes(
      List<int> **presizes,
      List<int> **postsizes
      )
{
    List<DataValueNode*>  *curRowList;
    int       numCols = 0;
    int       curCol = 0;
    string    curStr;
    int       curPre;
    int       curPost;
    int       curIdx;
    int       dotIdx;
    char      *tmpStr;

    if( myIters.Length() > 0 )
    {
	myIters.Reset();
	if( myIters.Current() != NULL )
	{
	    numCols = myIters.Current()->myVals.Length();
	}
    }
    *presizes = new List<int>;
    *postsizes= new List<int>;
    
    // Make the starting stub column info for each column:
    for( curCol = 0 ; curCol < numCols ; curCol++ )
    {
	(*presizes)->AddToEnd( 0 );
	(*postsizes)->AddToEnd( -1 );
    }
    if( numCols == 0 )
	return;  // give up - nothing here;

    // For each row in the table:
    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
    {
	curRowList = &( myIters.Current()->myVals );
	(*presizes)->Reset();
	(*postsizes)->Reset();
	// For each value in this row:
	for( curRowList->Reset() ; ! curRowList->AtEnd() ; curRowList->Next() )
	{
	    curStr = curRowList->Current()->myValue();
	    curPre = curRowList->Current()->myLongestStr();

	    curPost = -1;
	    // The numeric check:  Is it even eligable to be a number?
	    if( curRowList->Current()->myDelimType() == DataValueNode::NON )
	    {
		tmpStr = strdup( curStr.c_str() );
		// Two things in one - checking for valid syntax and
		// finding the decimal point.
		// -------------------------------------------------
	
		curIdx = 0;
		dotIdx = -1;
		// Consume leading whitespace:
		while( tmpStr[curIdx] == ' ' || tmpStr[curIdx] == '\t' )
		    curIdx++;
		// Consume optional pos/neg sign:
		if( tmpStr[curIdx] == '-' || tmpStr[curIdx] == '+' )
		    curIdx++;
		// Now a string of digits ended by a dot or whitespace or
		// the end of the string.  Anything else is an indicator
		// that this is not a true numeric value:
		while( isdigit( tmpStr[curIdx] ) )
		    curIdx++;
		curPre = curIdx+1;
		if( tmpStr[curIdx] == '.' )
		{   dotIdx = curIdx;
		    curIdx++;
		}
		curPost = 0;
		while( isdigit( tmpStr[curIdx] ) )
		{
		    curPost++; // count digits after the dot.
		    curIdx++;
		}
		while( isspace( tmpStr[curIdx] ) )  // trailing whitespace
		    curIdx++;

		// Okay, so the check is: if it stopped because it hit the
		// end of the string - the string was fully numeric, else
		// it had other garbage in it.
		if( tmpStr[curIdx] != '\0' )
		{
		    // Not numeric after all, so reset to the values
		    // for text strings:
		    curPost = -1;
	            curPre = curRowList->Current()->myLongestStr();
		}
		free(tmpStr);
	    }

	    if( (*presizes)->Current() < curPre )
		(*presizes)->AlterCurrent( curPre );
	    
	    if( (*postsizes)->Current() < curPost )
		(*postsizes)->AlterCurrent( curPost );
	    
	    (*presizes)->Next();
	    (*postsizes)->Next();
	}
    }
}

DataLoopNode::DataLoopNode(DataLoopNameListNode *d, DataLoopValListNode *l,
			   const string &tflag)
: 
  myDefList(d),
  myValList(tflag)
{
  if (transform(l) != StarFileNode::OK)
    cerr << "Starlib error: # of values in loop does not match # of names."
	 << endl;
  delete l;

  if( myDefList != NULL )
      myDefList->setParent(this);
  myValList.setParent(this);
}

DataLoopNode::DataLoopNode(const string &tflag) : myValList(tflag)
{
  myDefList = new DataLoopNameListNode;

  if( myDefList != NULL )
      myDefList->setParent(this);
  myValList.setParent(this);
}

DataLoopNode::~DataLoopNode()
{
    amDeleting = true;
    unlinkMe();

    if( myDefList != NULL )
	delete myDefList;
    amDeleting = false;
}

void DataLoopNode::reset()
{
  myValList.reset();
}

///////////////////////////////////////////////////////////////////////
//Copy Constructor
///////////////////////////////////////////////////////////////////////

DataLoopNode::DataLoopNode( DataLoopNode& D) : myValList( D.myValList )
{
    (void)copyFrom( D, false );
}

DataLoopNode::DataLoopNode( bool link, DataLoopNode& D) : myValList( link, D.myValList )
{
    (void)copyFrom( D, link );
}

bool DataLoopNode::copyFrom( DataLoopNode &fromThis, bool linkWith )
{

    ASTnode::copyFrom( fromThis );
    myDefList = new DataLoopNameListNode( linkWith, *(fromThis.myDefList));

    if( myDefList != NULL )
      myDefList->setParent(this);
    myValList.setParent(this);

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

int DataLoopNode::getMaxDepth( void )
{
    return myDefList->size() - 1;
}

////////////////////////////////////////////////////////////////////////

List<ASTnode*> * DataLoopNode::searchByTag( string &searchFor )
{
    if( myDefList == NULL )
	return new List<ASTnode*>;  // return empty list
    else
	return  myDefList->searchByTag( searchFor );
}

List<ASTnode*> * DataLoopNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *matches;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( myDefList != NULL )
    {
	matches = myDefList->searchForType( type, delim );
	if( matches->Length() > 0 )
	    retVal->AddToEnd( *matches );
	delete matches;
    }

    matches = myValList.searchForType( type, delim );
    if( matches->Length() > 0 )
	retVal->AddToEnd( *matches );
    delete matches;

    return retVal;
}

bool DataLoopNode::unlinkChild( ASTnode *child )
{
    if( myDefList == child )
    {   myDefList = (DataLoopNameListNode*)NULL;
	return true;
    }
    else if( &myValList == child )
    {   
	// TODO: Fix this.
	return false;  // Not sure how to go about deleting
	               // something that is not a pointer.
	               // myValList is a real instantiation,
	               // not a pointer.  Unfortunately, its
	               // type, LoopTableNode, has no 'clearall'
	               // method, so this would be ugly.
    }
    else
	return false;
}

List<ASTnode*> * DataLoopNode::searchByTagValue( string &tag,
	                                  string &value )
{
    int                   nest = -1;
    int                   col  = -1;
    ASTlist<DataValueNode*>  *colValues;
    List<ASTnode*>        *retVal = new List<ASTnode*>;

    tagPositionDeep( tag, &nest, &col );

    if( nest >= 0 )
    {
	// TODO: DELETE THIS NEXT PART IF AND WHEN NESTED
	// LOOPS ARE IMPLEMENTED FOR THIS FUNCTION!!!!!
	// ----------------------------------------------
	if( nest < 1 )
	{
	    // THIS ONLY WORKS ON NON-NESTED LOOPS!!!  I NEED TO
	    // MAKE THIS MORE SOPHISTICATED WHEN I HAVE THE TIME!
	    // --------------------------------------------------
	    colValues = myValList.returnLoopValues( col );

	    for( colValues->Reset() ; ! colValues->AtEnd() ; colValues->Next() )
	    {   if( *(colValues->Current()) == value )
		{   retVal->AddToEnd( colValues->Current() );
		}
	    }
	}
    }
    return retVal;
}


StarFileNode::Status DataLoopNode::transform(DataLoopValListNode *v)
{
  DataValueNode *value;
  DataLoopNameListNode::Status dstat;
  DataLoopValListNode::Status vstat;
  int row;

  v->reset();
  myDefList->reset();

  while ((vstat = v->getNextValue(value)) != DataLoopValListNode::END) {
    dstat = myDefList->getNextTag(vstat, row);
    if (dstat == DataLoopNameListNode::ERROR)
      return StarFileNode::ERROR;

    if( row >= 0 ) // row is negative only when the very last STOP
	           // for the DataLoopNode has been encountered,
		   // the negative value indicating that the nestlevel
		   // has popped all the way back past the starting point.
    {
	myValList.addValue(dstat, row, 0, value);
    }
  }
  return StarFileNode::OK;
}

SaveFrameListNode::SaveFrameListNode(ASTlist<DataNode *> *L) : myKids(L)
{
    myKids->setParent(this);
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        myKids->Current()->setParent(this);
}

SaveFrameListNode::SaveFrameListNode()
{
  myKids = new ASTlist<DataNode *>;
  myKids->setParent(this);
  for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
      myKids->Current()->setParent(this);
}

SaveFrameListNode::SaveFrameListNode( SaveFrameListNode &L )
{
    (void)copyFrom( L, false );
}

SaveFrameListNode::SaveFrameListNode( bool link, SaveFrameListNode &L )
{
    (void)copyFrom( L, link );
}

bool SaveFrameListNode::copyFrom( SaveFrameListNode &fromThis, bool linkWith )
{
    SaveFrameListNode &L = fromThis;

    ASTnode::copyFrom( fromThis );
    DataNode *remember = (L.myKids->AtEnd()) ? 
			(DataNode*)NULL : L.myKids->Current();
    myKids = new ASTlist<DataNode *>;
    myKids->setParent(this);
    for( L.myKids->Reset() ; ! L.myKids->AtEnd() ; L.myKids->Next() )
    {
	switch( L.myKids->Current()->myType() )
	{
	    case DataNode::DATAITEMNODE:
		myKids->AddToEnd( new DataItemNode( linkWith, *( (DataItemNode*)L.myKids->Current() ) )  );
		break;
	    case DataNode::DATALOOPNODE:
		myKids->AddToEnd( new DataLoopNode( linkWith, *( (DataLoopNode*)L.myKids->Current() ) )  );
		break;
	    case DataNode::SAVEFRAMENODE:
		myKids->AddToEnd( new SaveFrameNode( linkWith, *( (SaveFrameNode*)L.myKids->Current() ) )  );
		break;
	    default:   // gets gcc -Wall to shut up.
		break;
	}
    }
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
        myKids->Current()->setParent(this);

    (void)L.myKids->SeekTo( remember );

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

SaveFrameListNode::~SaveFrameListNode()
{
    amDeleting = true;
    unlinkMe();
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
	if( myKids->Current() != NULL )
	    delete myKids->Current();

    if( myKids != NULL )
	delete myKids;
    amDeleting = false;
}

List<ASTnode*> * SaveFrameListNode::searchByTag( string &searchFor )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchByTag( searchFor );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }
    return retVal;
}

List<ASTnode*> * SaveFrameListNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }
    return retVal;
}

bool SaveFrameListNode::unlinkChild( ASTnode *child )
{
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {   if( myKids->Current() == child )
	{   myKids->RemoveCurrent();
	    return true;
 	}
    }
    return false;
}

List<ASTnode*> * SaveFrameListNode::searchByTagValue( string &tag,
	                                       string &value )
{
    List<ASTnode*> *retVal = new List<ASTnode*>;
    List<ASTnode*> *foundSome = new List<ASTnode*>;

    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	foundSome = myKids->Current()->searchByTagValue( tag, value );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }
    return retVal;
}

int SaveFrameListNode::myLongestStr( void )
{
    int retVal = 0;
    int curLen = 0;

    // Return the longest of the 'myLongestStr()'s from the
    // list of DataNodes in this object:
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	// Loops don't count - they're weird.
	if( ! myKids->Current()->isOfType( ASTnode::DATALOOPNODE )  )
	{
	    curLen = myKids->Current()->myLongestStr();
	    if( curLen > retVal )
		retVal = curLen;
	}
    }
    return retVal;
}



SaveFrameNode::SaveFrameNode(SaveHeadingNode *h, SaveFrameListNode *l)
: 
  myHeading(h), mySaveFrameList(l)
{
    if( myHeading != NULL)
	myHeading->setParent(this);
    if( mySaveFrameList != NULL)
	mySaveFrameList->setParent(this);
}

SaveFrameNode::SaveFrameNode(const string & name)
{
  myHeading = new SaveHeadingNode(name);
  mySaveFrameList = new SaveFrameListNode;

  if( myHeading != NULL)
      myHeading->setParent(this);
  if( mySaveFrameList != NULL)
      mySaveFrameList->setParent(this);
}

SaveFrameNode::SaveFrameNode( SaveFrameNode &S )
{
    (void)copyFrom( S, false );
}

SaveFrameNode::SaveFrameNode( bool link, SaveFrameNode &S )
{
    (void)copyFrom( S, link );
}

bool SaveFrameNode::copyFrom( SaveFrameNode &fromThis, bool linkWith )
{
    ASTnode::copyFrom( fromThis );
    mySaveFrameList = new SaveFrameListNode( linkWith, *(fromThis.mySaveFrameList) );
    myHeading = new SaveHeadingNode( fromThis.myHeading->myName() );
    if( linkWith )
    {
	myHeading->setPeer( fromThis.myHeading );
	fromThis.myHeading->setPeer( myHeading );
    }

    if( myHeading != NULL)
	myHeading->setParent(this);
    if( mySaveFrameList != NULL)
	mySaveFrameList->setParent(this);

    if( linkWith )
    {   setPeer( &fromThis );
	fromThis.setPeer( this );
    }
    return true;
}

SaveFrameNode::~SaveFrameNode()
{
    amDeleting = true;
    unlinkMe();
    if( myHeading != NULL )
	delete myHeading;
    if( mySaveFrameList != NULL )
	delete mySaveFrameList;
    amDeleting = false;
}

List<ASTnode*> * SaveFrameNode::searchByTag( string &searchFor )
{
    List <ASTnode*> *retVal = new List<ASTnode*>;
    List <ASTnode*> *foundSome;

    if( myHeading->myName() == searchFor )
	retVal->AddToEnd( this );

    if( mySaveFrameList != (ASTnode*)NULL )
    {
	foundSome = mySaveFrameList->searchByTag( searchFor );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

List<ASTnode*> * SaveFrameNode::searchForType(
	ASTtype   type,       // The type to search for.
	int       delim
	)
{
    List <ASTnode*> *retVal = new List<ASTnode*>;
    List <ASTnode*> *foundSome;

    if( isOfType( type ) )
	retVal->AddToEnd( this );

    if( myHeading != NULL )
    {
	foundSome = myHeading->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }
    if( mySaveFrameList != (ASTnode*)NULL )
    {
	foundSome = mySaveFrameList->searchForType( type, delim );
	if( foundSome->Length() > 0 )
	    retVal->AddToEnd( *foundSome );
	delete foundSome;
    }

    return retVal;
}

bool SaveFrameNode::unlinkChild( ASTnode *child )
{
    if( myHeading == (SaveHeadingNode*)child )
    {   myHeading = (SaveHeadingNode*)NULL;
	return true;
    }
    else if( mySaveFrameList == (SaveFrameListNode*)child )
    {   mySaveFrameList = (SaveFrameListNode*)NULL;
	return true;
    }
    else
	return false;
}

List<ASTnode*> * SaveFrameNode::searchByTagValue( string &tag,
	                                   string &value )
{
    if( mySaveFrameList != (ASTnode*)NULL )
	return  mySaveFrameList->searchByTagValue( tag, value );
    else
	return new List<ASTnode*>;
}

void SaveFrameNode::changeName(string &name)
{
    myHeading->changeName( name );
}

SaveHeadingNode::SaveHeadingNode(const string & myHeading) 
: 
  HeadingNode(myHeading)
{}

SaveHeadingNode::SaveHeadingNode( SaveHeadingNode &N)
    : HeadingNode( N )
{}

SaveHeadingNode::SaveHeadingNode( bool link, SaveHeadingNode &N)
    : HeadingNode( link, N )
{}


DataHeadingNode::DataHeadingNode(const string & myHeading) 
: 
  HeadingNode(myHeading)
{}

DataHeadingNode::DataHeadingNode( DataHeadingNode &N)
    : HeadingNode( N )
{}

DataHeadingNode::DataHeadingNode( bool link, DataHeadingNode &N)
    : HeadingNode( link, N )
{}

GlobalHeadingNode::GlobalHeadingNode(const string & myHeading) 
: 
  HeadingNode(myHeading)
{}

GlobalHeadingNode::GlobalHeadingNode( GlobalHeadingNode &N)
    : HeadingNode( N )
{}

GlobalHeadingNode::GlobalHeadingNode( bool link, GlobalHeadingNode &N)
    : HeadingNode( link, N )
{}


DataBlockNode::DataBlockNode(const string & myHeading)
: 
  BlockNode(myHeading)
{}

DataBlockNode::DataBlockNode(HeadingNode *h, DataListNode *n) 
: 
  BlockNode(h, n)
{}

DataBlockNode::DataBlockNode(DataBlockNode &n)
    : BlockNode( n )
{}

DataBlockNode::DataBlockNode( bool link, DataBlockNode &n)
    : BlockNode( link, n )
{}


GlobalBlockNode::GlobalBlockNode(HeadingNode *h, DataListNode *n) 
    : BlockNode(h, n)
{}

GlobalBlockNode::GlobalBlockNode(GlobalBlockNode &n)
    : BlockNode( n )
{}

GlobalBlockNode::GlobalBlockNode( bool link, GlobalBlockNode &n)
    : BlockNode( link, n )
{}



////////////////////////////////////////////////////////////////////
////////    Here begins the text that used to be in the     ////////
////////    file called "astoper.cc".  I have left in the   ////////
////////    RCS log for the sake of posterity, but it is    ////////
////////    no longer updated on each check-in.  Only the   ////////
////////    main RCS log at the top of this file is updated ////////
////////    now.  There was absolutely zero logical reason  ////////
////////    to split up these routines the way they were    ////////
////////    being split into different files.   Splitting   ////////
////////    things up is good, but in this case they were   ////////
////////    only split based on who was coding which pieces ////////
////////    rather than on any modular logical breakdown.   ////////
////////////////////////////////////////////////////////////////////

//////////////////////////// RCS COMMENTS: /////////////////////////////
//  HISTORY LOG:
//  ------------
//
// Revision 1.4  1998/06/01  21:15:03  madings
// Changed FlattenNestedLoop so it returns a List<> instead of an ASTlist<>.
// ASTlist<> should NEVER, EVER, EVER be used to hold a list that is outside
// the normal star tree.  (ASTlist<> changes the parent pointers of the
// nodes you add to it, so if you use it to hold pointers to nodes in a
// star tree, you change the parent pointers of those nodes.)
//
// Revision 1.3  1998/06/01  18:58:30  madings
// Changed FlattenNestedLoops() so that it returns a list of pointers
// to the names and values that are actually in the star tree, as
// opposed to generating new copies of those nodes and making a list
// of the new copies.
//
// Revision 1.2  1998/05/08  20:27:24  informix
// Made a change to handle the C++ STL instead of gcc's libg++.
// Now all strings are of type 'string' instead of 'String' (notice
// the uppercase/lowercase change.)  Applications using this library
// will have to change also, unfortunately, to use the new string style.
//
// Revision 1.1.1.1  1998/02/12  20:59:00  informix
// re-creation of starlib project, without sb_lib stuff.  sb_lib has
// been moved to it's own project.
//
// Revision 1.1.1.1  1997/11/07  01:49:07  madings
// First check-in of starlib and sb_lib in library form.
//
// Revision 1.3  1997/10/10  18:34:22  madings
// Changed DataValueNode so that it now no longer has subclasses for
// single-quote-delimited strings, double-quote delimited strings,
// semicolon-delimited strings, and so on.  Now there is only one kind
// of DataValueNode, and it uses a flag to decide on the delimiter type.
// This also allows for the creation of methods to change a DataValueNode
// in place, which was not previously possible when different kinds of
// values had to be stored in different kinds of classes.
//
// Revision 1.2  1997/10/01  22:35:03  bsri
// This version of starlib split main.cc into separate files (which contain
// the transformation functions - one in each file).
//
// Revision 1.1.1.1  1997/09/12  22:54:40  madings
// Restarting CVS tree because I lost the CVS dir.
//
//Revision 1.6  1997/07/30  21:34:26  madings
//check-in before switching to cvs
//
//Revision 1.5  1997/07/11  22:40:03  madings
//Finally seems to work - just waiting for
//new dictionaries.
//
//Revision 1.4  1997/06/13  22:52:44  madings
//Checking in before bringing in sridevi code.
//
//Revision 1.3  1997/06/04  18:57:57  madings
//The RCS ID variables were missing the trailing $, so I
//changed them across all these files.
//
//Revision 1.2  1997/06/04  18:52:44  madings
//I Added the RCS log and ID variables.
//
//
////////////////////////////////////////////////////////////////////////


//#############################################################################
//#############################################################################
//#############################################################################
//Given the name of a  data item in a SAVEFRAME object this
//function returns a pointer to the data item which can ONLY be
//a loop or simple data item 


DataNode*  SaveFrameNode::ReturnSaveFrameDataNode(string dataItemName)
{
  return (mySaveFrameList->ReturnSaveFrameDataNode(dataItemName));
}


DataNode*  SaveFrameListNode::ReturnSaveFrameDataNode(string dataItemName)
{
  myKids->Reset();
  
  while(!myKids->AtEnd())
    {
      if (myKids->Current()->myName() == dataItemName)
	  return (myKids->Current());

      myKids->Next();
    }
  return NULL;
}

DataNode* SaveFrameNode::ReturnSaveFrameDeepLook(  string dataItemName, int *nest, int *col )
{
    // Defaults:
    (*nest) = -1;
    (*col) = -1;

    return mySaveFrameList->ReturnSaveFrameDeepLook( dataItemName, nest, col );
}

DataNode* SaveFrameListNode::ReturnSaveFrameDeepLook( string dataItemName, int *nest, int *col )
{
    myKids->Reset(); 
    while(!myKids->AtEnd())
    {
	//                      ITEM TYPE:
	if( myKids->Current()->myType() == DataNode::DATAITEMNODE )
	{
	    if (myKids->Current()->myName() == dataItemName)
		return myKids->Current();
	}
	//                      LOOP TYPE:
	else if( myKids->Current()->myType() == DataNode::DATALOOPNODE )
	{
	    ( (DataLoopNode*)myKids->Current() )->tagPositionDeep( dataItemName, nest, col );

	    if( (*nest) >= 0 )
		return myKids->Current();
	}

	myKids->Next();
    }
    return NULL;
}


//#############################################################################
//#############################################################################
//#############################################################################


//The functions below returns a pointer to a DataNode in a specified
//datablock and a specified data item name in the data block.
//Note this item could be a simple data item, a saveframe object
//or a loop.NULL is returned if the item is not present.


DataNode* StarFileNode::  ReturnDataBlockDataNode(
						  string datablockName, 
						  string dataitemName
						  )
{
  return (myStarList->ReturnDataBlockDataNode(datablockName, dataitemName));
}




DataNode* StarListNode::  ReturnDataBlockDataNode(
						  string datablockName,
						  string dataitemName
						  )
{

//Checks to see if there is a match in the data block
  myKids->Reset();
  while(!myKids->AtEnd())
    {
      if(myKids->Current()->myName() == datablockName)
	return (myKids->Current()->ReturnDataBlockDataNode(dataitemName));
      myKids->Next();
    }
  return NULL;
}



DataNode* BlockNode:: ReturnDataBlockDataNode(string dataitemName)
{
  return (myDataList->ReturnDataBlockDataNode(dataitemName));
}



DataNode* DataListNode:: ReturnDataBlockDataNode(string dataitemName)
  {
    myKids->Reset();
    while(!myKids->AtEnd())
      {
	if( (myKids->Current()->myName() == dataitemName))
	  return (myKids->Current());
	myKids->Next();
      }
    return NULL;
  }

void DataLoopNode::tagPositionDeep( string tagName, int *nestLevel, int *column )
{
    myDefList->tagPositionDeep( tagName, nestLevel, column ) ;
}

void DataLoopNameListNode::tagPositionDeep( string tagName, int *nestLevel, int *column )
{
    int pos = -1;
    for( myKids->Reset() , (*nestLevel) = 0 ; ! myKids->AtEnd() ; myKids->Next() , (*nestLevel)++ )
    {
	pos = myKids->Current()->tagPosition( tagName );

	if( pos >= 0)
	    break;
    }
    if( pos >= 0 )
    {   (*column) = pos;
	return;
    }
    else
    {   (*column) = -1;
	(*nestLevel) = -1;
	return;
    }
}

//############################################################################
//############################################################################
//############################################################################
//These set of functions will perform remove a specific saveframe
//from a specified data block. No action will be taken if either
//does not exist in the tree.

void  StarFileNode::  RemoveSaveFrame(
				      string datablockName, 
				      string saveframeName
				      )
{
  myStarList->RemoveSaveFrame(datablockName, saveframeName);
}



void  StarListNode::RemoveSaveFrame(
				    string datablockName, 
				    string saveframeName
				    )
{
  //Checks to see if there is a match in the data block
  myKids->Reset();
  bool done = false;
  while(!myKids->AtEnd() && !done)
    {
      if(myKids->Current()->myName() == datablockName)
	{
	  myKids->Current()->RemoveSaveFrame(saveframeName);
	  done = true;
	}
      else
	myKids->Next();
    }
}


void BlockNode::RemoveSaveFrame(string saveframeName)
{
  
    myDataList->RemoveSaveFrame(saveframeName);
}


void  DataListNode:: RemoveSaveFrame(string saveframeName)
  {
    myKids->Reset();
    bool done = false;
    while(!myKids->AtEnd())
      {
	  if( (myKids->Current()->myName() == saveframeName) && DataNode::SAVEFRAMENODE ==myKids->Current()->myType() && !done)
	  {
	    myKids->RemoveCurrent();
	    done = true;
	  }
	else
	  myKids->Next();
      }
  }
//########################################################################
//########################################################################
//########################################################################
//These functions will add a given data item to a specified saveframe
void SaveFrameNode::AddItemToSaveFrame(DataNode* S)
{
  mySaveFrameList->AddItemToSaveFrame(S);
}


void SaveFrameListNode::AddItemToSaveFrame(DataNode* S)
{
  S->setParent(this);
  myKids->AddToEnd(S);
}


//#########################################################################
//#########################################################################
//#########################################################################
//These functions will add a given data item to a specified datablock

void StarFileNode::AddSaveFrameToDataBlock(
					   string datablockName, 
					   DataNode* newNode
					   )
{
   
   myStarList->AddSaveFrameToDataBlock(datablockName,newNode);
}


void StarListNode::AddSaveFrameToDataBlock(
					   string datablockName, 
					   DataNode* newNode
					   )
{

  myKids->Reset();
  while(!myKids->AtEnd())
    {
      if(myKids->Current()->myName() == datablockName)
	{
	 
	  myKids->Current()->AddSaveFrameToDataBlock(newNode);
	  
	  return;
	}
      myKids->Next();
    }

}


void  BlockNode:: AddSaveFrameToDataBlock(DataNode* newNode)
{
   
  myDataList->AddSaveFrameToDataBlock(newNode);
}


void DataListNode:: AddSaveFrameToDataBlock(DataNode* newNode)
{

  newNode->setParent( this );
  myKids->AddToEnd(newNode);
  
}
//############################################################################
//############################################################################
//############################################################################
//The following function returns a pointer to the list of dataitems
//contained in a datablock


ASTlist<DataNode*>* BlockNode::GiveMyDataList()
{
  return(myDataList->GiveMyDataList()); 
}


ASTlist<DataNode*>* DataListNode::GiveMyDataList()
{
  return (myKids);
}



//#############################################################################
//#############################################################################
//#############################################################################

//The following function returns a pointer to the list of dataitems
//contained in a saveframe

ASTlist<DataNode*>* SaveFrameNode::GiveMyDataList()
{
  return(mySaveFrameList->GiveMyDataList());
}


ASTlist<DataNode*>* SaveFrameListNode::GiveMyDataList()
{
  return (myKids);
}
//############################################################################
//###########################################################################
//##########################################################################
//The following functions give a pointer to the list of datablocks in the AST
ASTlist<BlockNode*>* StarFileNode::GiveMyDataBlockList()
{
  return (myStarList->GiveMyDataBlockList());
}


ASTlist<BlockNode*>* StarListNode::GiveMyDataBlockList()
{
  return myKids;
}

//##########################################################################
//##########################################################################
//##########################################################################
//The following functions are responsible for flattening nested
//loop structures into linear lists of tags and values

void DataLoopNode::FlattenNestedLoop(
				     List<DataNameNode*> * &L,
				     List<DataValueNode*>* &M
				     )
				    
{
  L = new List<DataNameNode *>;
  M = new List<DataValueNode*>;

  myDefList->FlattenNestedLoop(L);
  myValList.FlattenNestedLoop(M);
}



void DataLoopNameListNode::FlattenNestedLoop(List<DataNameNode*>*  L)
{
  
  
  myKids->Reset();
  while(!myKids->AtEnd())
   {
      myKids->Current()->FlattenNestedLoop(L);
      myKids->Next();
    }
}



void LoopNameListNode::FlattenNestedLoop(List<DataNameNode*>* L)
{
  myKids->Reset();
  while(!myKids->AtEnd())
    {
      DataNameNode* tempNode = myKids->Current();
      L->AddToEnd(tempNode);
      myKids->Next();
    }
}



void LoopTableNode::FlattenNestedLoop(List<DataValueNode*>* M)
{
  myIters.Reset();
  while(!myIters.AtEnd())
    {
      myIters.Current()->FlattenNestedLoop(M);
      myIters.Next();
    }
}



void LoopRowNode::FlattenNestedLoop(List<DataValueNode*>* M)
{
  myVals.Reset();
  while(!myVals.AtEnd())
    {
      DataValueNode* tempNode = myVals.Current();
      M->AddToEnd(tempNode);
      myVals.Next();
    }
  if(innerLoop)
    innerLoop->FlattenNestedLoop(M);
}
////////////////////////////////////////////////
//For part 2
///////////////////////////////////////////////
void LoopRowNode::reset()
{
  myVals.Reset();
  if (innerLoop != NULL)
    innerLoop->reset();
}

DataLoopValListNode::Status LoopRowNode::returnNextLoopElement(DataValueNode* & v)
{
  if (!myVals.AtEnd()) {
    v = myVals.Current();
    myVals.Next();
    return DataLoopValListNode::CONTINUE;
  }

  if (innerLoop)
    return innerLoop->returnNextLoopElement(v);
  else
    return DataLoopValListNode::END;
}

DataValueNode* LoopRowNode::returnLoopValue(unsigned position)
{
  unsigned i = 0;

  for (myVals.Reset(); !myVals.AtEnd(); myVals.Next(), i++)
    if (i == position)
      return myVals.Current();

  return NULL;
}

void LoopTableNode::reset()
{
  for (myIters.Reset(); !myIters.AtEnd(); myIters.Next())
    myIters.Current()->reset();
  myIters.Reset();
}

DataLoopValListNode::Status LoopTableNode::returnNextLoopElement(DataValueNode* & v)
{
  // If we are at the end the first time, we will return STOP. This is done
  // by the last statement in this function. Otherwise, we have already
  // returned STOP once, now we have to return END.
  // The distinction between STOP and END is the following :
  // STOP means that some loop has completed, not necessarily this one, it may
  // be some inner loop. That information is reported to the caller.
  // END means that this loop and all its children do not have any more data.

  if (myIters.AtEnd())
    return DataLoopValListNode::END;

  while (!myIters.AtEnd()) {

    DataLoopValListNode::Status retVal = 
      myIters.Current()->returnNextLoopElement(v);

    if (retVal == DataLoopValListNode::STOP)
      return DataLoopValListNode::STOP;

    if (retVal == DataLoopValListNode::END)
      myIters.Next();
    else
      return DataLoopValListNode::CONTINUE;
  }    

  return DataLoopValListNode::STOP;
}

ASTlist<DataValueNode *> * LoopTableNode::returnLoopValues(unsigned position)
{
  ASTlist<DataValueNode *> *retVal = new ASTlist<DataValueNode *>;

  retVal->setParent(this);

  for (myIters.Reset(); !myIters.AtEnd(); myIters.Next())
    retVal->AddToEnd(myIters.Current()->returnLoopValue(position));

  return retVal;
}

DataLoopValListNode::Status DataLoopNode::returnNextLoopElement(
						    DataValueNode* & v)
{
  return myValList.returnNextLoopElement(v);
}

ASTlist<DataValueNode *> * DataLoopNode::returnLoopValues(const string & tagName)
{
  // Caution : This function only searches in the first level of loop
  int tagPos;

  if ((tagPos = myDefList->tagPosition(tagName)) == -1)
    // Not found
    return NULL;

  return myValList.returnLoopValues(tagPos);
}

int DataLoopNameListNode::tagPosition(const string & tagName)
{
  // Caution : This function only searches in the first level of loop
  myKids->Reset();
  return myKids->Current()->tagPosition(tagName);
}

int LoopNameListNode::tagPosition(const string & tagName)
{
  int i = 0;
  for (myKids->Reset(); !myKids->AtEnd(); myKids->Next(), i++)
  {
    if (tagName == myKids->Current()->myName())
    {
      return i;
    }
  }

  // Not found
  return -1;
}

ASTlist<DataValueNode *> * DataItemNode::returnLoopValues(const string & tagName)
{
  if (tagName != myName())
    return NULL;

  ASTlist<DataValueNode *> *retVal = new ASTlist<DataValueNode *>;
  retVal->setParent( myDataValue->myParent() );
  retVal->AddToEnd(myDataValue);

  return retVal;
}

ASTlist<DataValueNode *> * SaveFrameNode::returnLoopValues(const string & )
{
  return NULL;
}



StarFileNode::Status DataLoopNameListNode::RemoveDataName( const string & name,
	                                                  int   *nest,
							  int   *ele   )
{
    if( nest )
	*nest = 0;
    // For each LoopNameListNode in this DataLoopDefListNode:
    for(  myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next()  )
    {
	if( nest )
	    (*nest)++;
	// Try to remove the name from the list.  If it was found and
	// removed, then we are done, get out:
	if( myKids->Current()->RemoveDataName( name, ele ) == StarFileNode::OK )
	    return StarFileNode::OK;
    }
    if( nest )
	*nest = 0;
    return StarFileNode::ERROR;
}

StarFileNode::Status LoopNameListNode::RemoveDataName(const string & name,
	                                             int          * ele )
{
    if( ele )
	*ele = 0;
    for( myKids->Reset() ; ! myKids->AtEnd() ; myKids->Next() )
    {
	if( ele )
	    (*ele)++;
	if( myKids->Current()->myName() == name )
	{
	    myKids->RemoveCurrent();
	    return StarFileNode::OK;  // found it - okay, we're done
	}
    }
    if( ele )
	*ele = 0;
    return StarFileNode::ERROR;
}

StarFileNode::Status  DataLoopNode::ChangeName(
	const string &oldName,
	const string &newName
	)
{
    return myDefList->ChangeName( oldName, newName );
}

StarFileNode::Status  DataLoopNode::RemoveColumn( const string name )
{
    int  nestLevel, columnNum;
    StarFileNode::Status   retVal;

    retVal = myDefList->RemoveDataName( name, &nestLevel, &columnNum );

    if( retVal == StarFileNode::OK )
	myValList.RemoveColumnValues( nestLevel, columnNum );

    return retVal;
}


//   These next two functions form a recursive pair.  Watch their
//   intertwining carefully.  LoopTableNode's RemoveColumnValues() calls
//   IterLoop's RemoveColumnValues() for each row of the loop.  If
//   IterLoop's RemoveColumnValues() realizes that it is not at
//   the base case of recursion yet, then it calls the
//   RemoveColumnValues() of its inner LoopTableNode.  (Which will call
//   RemoveColumnValues() for each of its rows and so on.)

StarFileNode::Status  LoopRowNode::RemoveColumnValues(
	int          nest,
	int          ele )
{
    StarFileNode::Status  retVal = StarFileNode::OK;

    // Base case: delete the element from the list.
    if( nest == 0 )
    {
	// Count off until we have reached the number of the
	// element we want to chuck away:
	for( myVals.Reset() ; (! myVals.AtEnd()) && ele > 1 ; myVals.Next() , --ele )
	{ /*void-body*/ }

	if( ! myVals.AtEnd() )  // Chuck it
	    myVals.RemoveCurrent();
	else
	    retVal = StarFileNode::ERROR;
    }
    else // Recursive case - call it for my inner loop.
    {
	retVal = innerLoop->RemoveColumnValues( nest, ele );
    }

    return retVal;
}

StarFileNode::Status  LoopTableNode::RemoveColumnValues(
	int          nest,
	int          ele )
{
    StarFileNode::Status  retVal = StarFileNode::ERROR;
    for( myIters.Reset() ; ! myIters.AtEnd() ; myIters.Next() )
    {
	if( myIters.Current()->RemoveColumnValues( nest-1, ele ) == StarFileNode::OK )
	    retVal = StarFileNode::OK ;
    }

    return retVal;
}

StarFileNode::Status  DataLoopNameListNode::ChangeName(
	const string &oldName,
	const string &newName
	)
{
    StarFileNode::Status retVal = StarFileNode::ERROR;
    for(   myKids->Reset() ;
	   ! myKids->AtEnd() && retVal == StarFileNode::ERROR ;
	   myKids->Next()   )
    {
	retVal = myKids->Current()->ChangeName( oldName, newName );
    }
    return retVal;
}

StarFileNode::Status LoopNameListNode::ChangeName(
	const string &oldName,
	const string &newName
	)
{
    StarFileNode::Status retVal = StarFileNode::ERROR;
    for(   myKids->Reset() ;
	   ! myKids->AtEnd() && retVal == StarFileNode::ERROR ;
	   myKids->Next()   )
    {
	if( myKids->Current()->myName() == oldName )
	{
	    myKids->Current()->setName( newName );
	    retVal = StarFileNode::OK;
	}
    }
    return retVal;
}



////////////////////////////////////////////////////////////////////
////////    Here begins the text that used to be in the     ////////
////////    file called "interface.cc".  I have left in the ////////
////////    RCS log for the sake of posterity, but it is    ////////
////////    no longer updated on each check-in.  Only the   ////////
////////    main RCS log at the top of this file is updated ////////
////////    now.  There was absolutely zero logical reason  ////////
////////    to split up these routines the way they were    ////////
////////    being split into different files.   Splitting   ////////
////////    things up is good, but in this case they were   ////////
////////    only split based on who was coding which pieces ////////
////////    rather than on any modular logical breakdown.   ////////
////////////////////////////////////////////////////////////////////

//////////////////////////// RCS COMMENTS: /////////////////////////////
//  HISTORY LOG:
//  ------------
//
// Revision 1.2  1998/05/08  20:27:24  informix
// Made a change to handle the C++ STL instead of gcc's libg++.
// Now all strings are of type 'string' instead of 'String' (notice
// the uppercase/lowercase change.)  Applications using this library
// will have to change also, unfortunately, to use the new string style.
//
// Revision 1.1.1.1  1998/02/12  20:59:00  informix
// re-creation of starlib project, without sb_lib stuff.  sb_lib has
// been moved to it's own project.
//
// Revision 1.1.1.1  1997/11/07  01:49:06  madings
// First check-in of starlib and sb_lib in library form.
//
// Revision 1.2  1997/10/10  18:34:23  madings
// Changed DataValueNode so that it now no longer has subclasses for
// single-quote-delimited strings, double-quote delimited strings,
// semicolon-delimited strings, and so on.  Now there is only one kind
// of DataValueNode, and it uses a flag to decide on the delimiter type.
// This also allows for the creation of methods to change a DataValueNode
// in place, which was not previously possible when different kinds of
// values had to be stored in different kinds of classes.
//
// Revision 1.1.1.1  1997/09/12  22:54:40  madings
// Restarting CVS tree because I lost the CVS dir.
//
//Revision 1.6  1997/07/30  21:34:26  madings
//check-in before switching to cvs
//
//Revision 1.5  1997/07/11  22:40:04  madings
//Finally seems to work - just waiting for
//new dictionaries.
//
//Revision 1.4  1997/06/13  22:52:44  madings
//Checking in before bringing in sridevi code.
//
//Revision 1.3  1997/06/04  18:57:58  madings
//The RCS ID variables were missing the trailing $, so I
//changed them across all these files.
//
//Revision 1.2  1997/06/04  18:52:44  madings
//I Added the RCS log and ID variables.
//
//
////////////////////////////////////////////////////////////////////////

StarFileNode::Status StarFileNode::AddDataBlock(const string & name)
{
  return myStarList->AddDataBlock(name);
}

StarFileNode::Status StarFileNode::AddSaveFrame(const string & name)
{
  return myStarList->AddSaveFrame(name);
}

StarFileNode::Status StarFileNode:: AddSaveFrameDataItem( const string & name,
	                                                  const string & value,
							  DataValueNode::ValType type)
{
  return myStarList->AddSaveFrameDataItem(name, value, type);
}

StarFileNode::Status StarFileNode::AddSaveFrameLoop()
{
  return myStarList->AddSaveFrameLoop();
}

StarFileNode::Status StarFileNode::AddSaveFrameLoopDataName(const string & name)
{
  return myStarList->AddSaveFrameLoopDataName(name);
}

StarFileNode::Status StarFileNode::AddSaveFrameLoopDataValue( const string & value,
						              DataValueNode::ValType type )
{
  return myStarList->AddSaveFrameLoopDataValue( value, type );
}

StarFileNode::Status StarListNode::AddDataBlock(const string & name)
{
  DataBlockNode *db = new DataBlockNode(name);
  db->setParent(this);
  myKids->AddToEnd(db);
  return StarFileNode::OK;
}

StarFileNode::Status StarListNode::AddSaveFrame(const string & name)
{
  BlockNode *db = myKids->Last();

  if (db == NULL)
    return StarFileNode::ERROR;
  else
    return db->AddSaveFrame(name);
}

StarFileNode::Status StarListNode::AddSaveFrameDataItem(const string & name, 
							const string & value,
							DataValueNode::ValType type)
{
  BlockNode *db = myKids->Last();

  if (db == NULL)
    return StarFileNode::ERROR;
  else
    return db->AddSaveFrameDataItem(name, value, type);
}

StarFileNode::Status StarListNode::AddSaveFrameLoop()
{
  BlockNode *db = myKids->Last();

  if (db == NULL)
    return StarFileNode::ERROR;
  else
    return db->AddSaveFrameLoop();
}

StarFileNode::Status StarListNode::AddSaveFrameLoopDataName(const string & name)
{
  BlockNode *db = myKids->Last();

  if (db == NULL)
    return StarFileNode::ERROR;
  else
    return db->AddSaveFrameLoopDataName(name);

}

StarFileNode::Status StarListNode::AddSaveFrameLoopDataValue( const string & value,
						              DataValueNode::ValType type )
{
  BlockNode *db = myKids->Last();

  if (db == NULL)
    return StarFileNode::ERROR;
  else
    return db->AddSaveFrameLoopDataValue(value, type);
}

StarFileNode::Status BlockNode::AddSaveFrame(const string & name)
{
  return myDataList->AddSaveFrame(name);
}

StarFileNode::Status BlockNode::AddSaveFrameDataItem(const string & name,
						     const string & value,
						     DataValueNode::ValType type )
{
  return myDataList->AddSaveFrameDataItem(name, value, type);
}

StarFileNode::Status BlockNode::AddSaveFrameLoop()
{
  return myDataList->AddSaveFrameLoop();
}

StarFileNode::Status BlockNode::AddSaveFrameLoopDataName(const string & name)
{
  return myDataList->AddSaveFrameLoopDataName(name);
}

StarFileNode::Status BlockNode::AddSaveFrameLoopDataValue( const string & name,
						           DataValueNode::ValType type )
{
  return myDataList->AddSaveFrameLoopDataValue(name, type);
}

StarFileNode::Status DataListNode::AddSaveFrame(const string & name)
{
  SaveFrameNode *sf = new SaveFrameNode(name);
  sf->setParent(this);
  myKids->AddToEnd(sf);
  return StarFileNode::OK;
}

StarFileNode::Status DataListNode::AddSaveFrameDataItem(const string & name, 
							const string & value,
							DataValueNode::ValType type)
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::SAVEFRAMENODE)
    return StarFileNode::ERROR;
  else
    return ((SaveFrameNode *)dn)->AddDataItem(name, value, type);
}

StarFileNode::Status DataListNode::AddSaveFrameLoop()
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::SAVEFRAMENODE)
    return StarFileNode::ERROR;
  else
    return ((SaveFrameNode *)dn)->AddLoop();
}

StarFileNode::Status DataListNode::AddSaveFrameLoopDataName(const string & name)
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::SAVEFRAMENODE)
    return StarFileNode::ERROR;
  else
    return ((SaveFrameNode *)dn)->AddLoopDataName(name);
}

StarFileNode::Status DataListNode::AddSaveFrameLoopDataValue( const string & name,
						              DataValueNode::ValType type )
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::SAVEFRAMENODE)
    return StarFileNode::ERROR;
  else
    return ((SaveFrameNode *)dn)->AddLoopDataValue(name, type);
}

StarFileNode::Status SaveFrameNode::AddDataItem(const string & name, 
						const string & value,
						DataValueNode::ValType type)
{
  return mySaveFrameList->AddDataItem(name, value, type);
}

StarFileNode::Status SaveFrameNode::AddLoop(const string &tflag)
{
  return mySaveFrameList->AddLoop(tflag);
}

StarFileNode::Status SaveFrameNode::AddLoopDataName(const string & name)
{
  return mySaveFrameList->AddLoopDataName(name);
}

StarFileNode::Status SaveFrameNode::AddLoopDataValue( const string & value,
					              DataValueNode::ValType type )
{
  return mySaveFrameList->AddLoopDataValue(value, type);
}
  
StarFileNode::Status SaveFrameListNode::AddDataItem(const string & name, 
						    const string & value,
						    DataValueNode::ValType type)
{
  DataNode *dn = new DataItemNode(name, value, type);
  dn->setParent(this);
  myKids->AddToEnd(dn);
  return StarFileNode::OK;
}

StarFileNode::Status SaveFrameListNode::AddLoop(const string &tflag)
{
  DataNode *dn = new DataLoopNode(tflag);
  dn->setParent(this);
  myKids->AddToEnd(dn);
  return StarFileNode::OK;
}

StarFileNode::Status SaveFrameListNode::AddLoopDataName(const string & name)
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::DATALOOPNODE)
    return StarFileNode::ERROR;
  else
    return ((DataLoopNode *)dn)->AddDataName(name);
}

StarFileNode::Status SaveFrameListNode::AddLoopDataValue( const string & value,
					                  DataValueNode::ValType type )
{
  DataNode *dn = myKids->Last();

  if (!dn || dn->myType() != DataNode::DATALOOPNODE)
    return StarFileNode::ERROR;
  else
    return ((DataLoopNode *)dn)->AddDataValue(value, type);
}
  
StarFileNode::Status DataLoopNode::AddDataName(const string & name)
{
  if (myDefList->AddDataName(name) == StarFileNode::ERROR)
    return StarFileNode::ERROR;

  myDefList->reset(); // This is needed so that when AddDataValue is done,
  // we can scan from the beginning of the name list to figure correct position

  return StarFileNode::OK;
}

StarFileNode::Status DataLoopNode::AddDataValue( const string & value,
					         DataValueNode::ValType type )
{
  int row;

  const DataLoopNameListNode::Status dstat = 
    myDefList->getNextTag(DataLoopValListNode::CONTINUE, row);

  if (row != 0 || (dstat != DataLoopNameListNode::NEW_ITER && 
		   dstat != DataLoopNameListNode::SAME_ITER))
    return StarFileNode::ERROR;

  DataValueNode *newValue = new DataValueNode( value, type );
  myValList.addValue(dstat, row, 0u, newValue );

  return StarFileNode::OK;
}

StarFileNode::Status DataLoopNameListNode::AddDataName(const string & name)
{
  // The list already has been allocated, now have to check if it is empty
  if (myKids->Length() == 0) {
    myKids->AddToEnd(new LoopNameListNode);
    myKids->Last()->setParent(this);
    myKids->Reset();
  }
  return myKids->Current()->AddDataName(name);
}

StarFileNode::Status LoopNameListNode::AddDataName(const string & name)
{
  myKids->AddToEnd(new DataNameNode(name));
  return StarFileNode::OK;
}


// unparse.cc used to be a separately compiled file, but due to some
// problems porting to the egcs compiler, it had to be included into
// this file instead of compiled separately.
#include "unparse.cc"

